
import json
import numbers
import warnings
from collections.abc import Iterable
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib.pyplot as plt
from cadquery import (Assembly, Color, Compound, Plane, Workplane, exporters,
                      importers)
from cadquery.occ_impl import shapes
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

import paramak
from paramak.utils import (_replace, cut_solid, facet_wire, get_hash,
                           intersect_solid, plotly_trace, union_solid)


class Shape:
    """A shape object that represents a 3d volume and can have materials and
    neutronics tallies assigned. Shape objects are not intended to be used
    directly by the user but provide basic functionality for user-facing
    classes that inherit from Shape. Provides a .show attribute for rendering
    in Jupyter Lab

    Args:
        points (tuple of (float, float, float), optional): the x, y, z
            coordinates of points that make up the shape. Defaults to None.
        connection_type (str, optional): The type of connection between points.
            Possible values are "straight", "circle", "spline", "mixed".
            Defaults to "mixed".
        name (str, optional): the name of the shape, used in the graph legend
            by export_html. Defaults to None.
        color ((float, float, float [, float]), optional): The color to use
            when exporting as html graphs or png images. Can be in RGB or RGBA
            format with floats between 0 and 1. Defaults to (0.5, 0.5, 0.5).
        azimuth_placement_angle (iterable of floats or float, optional): the
            azimuth angle(s) used when positioning the shape. If a list of
            angles is provided, the shape is duplicated at all angles.
            Defaults to 0.0.
        workplane: the orientation of the Cadquery workplane. Options include
            strings "XY", "YZ", "XZ" or a Cadquery.Plane(). Defaults to "XZ".
        rotation_axis (str or list, optional): rotation axis around which the
            solid is rotated. If None, the rotation axis will depend on the
            workplane or path_workplane if applicable. Can be set to "X", "-Y",
            "Z", etc. A custom axis can be set by setting a list of two XYZ
            floats. Defaults to None.
        tet_mesh (str, optional): If not None, a tet mesh flag will be added to
            the neutronics description output. Defaults to None.
        scale (float, optional): If not None, a scale flag will be added to
            the neutronics description output. Defaults to None.
        surface_reflectivity (Boolean, optional): If True, a
            surface_reflectivity flag will be added to the neutronics
            description output. Defaults to None.
        physical_groups (dict, optional): contains information on physical
            groups (volumes and surfaces). Defaults to None.
        cut (paramak.shape or list, optional): If set, the current solid will
            be cut with the provided solid or iterable in cut. Defaults to
            None.
        intersect (paramak.shape or list, optional): If set, the current solid
            will be interested with the provided solid or iterable of solids.
            Defaults to None.
        union (paramak.shape or list, optional): If set, the current solid
            will be united with the provided solid or iterable of solids.
            Defaults to None.
        graveyard_size: The dimension of cube shaped the graveyard region used
            by DAGMC. This attribute is used preferentially over
            graveyard_offset.
        graveyard_offset: The distance between the graveyard and the largest
            shape. If graveyard_size is set the this is ignored.
    """

    def __init__(
        self,
        points: Union[tuple, list] = None,
        connection_type: Optional[str] = "mixed",
        name: Optional[str] = None,
        color: Optional[Tuple[float, float, float,
                              Optional[float]]] = (0.5, 0.5, 0.5),
        azimuth_placement_angle: Optional[Union[float, List[float]]] = 0.0,
        workplane: Optional[Union[str, Plane]] = "XZ",
        rotation_axis: Optional[str] = None,
        tet_mesh: Optional[str] = None,
        scale: Optional[float] = None,
        surface_reflectivity: Optional[bool] = False,
        physical_groups=None,
        # TODO defining Shape types as paramak.Shape results in circular import
        cut=None,
        intersect=None,
        union=None,
        graveyard_size: Optional[float] = 20_000,
        graveyard_offset: Optional[float] = None,
    ):

        self.connection_type = connection_type
        self.points = points
        self.color = color
        self.name = name

        self.cut = cut
        self.intersect = intersect
        self.union = union

        self.azimuth_placement_angle = azimuth_placement_angle
        self.workplane = workplane
        self.rotation_axis = rotation_axis

        # initialise to something different than self.points
        # old_points is used in the processed_points getter
        self.old_points = 0

        # neutronics specific properties
        self.tet_mesh = tet_mesh
        self.scale = scale
        self.surface_reflectivity = surface_reflectivity
        self.graveyard_offset = graveyard_offset
        self.graveyard_size = graveyard_size

        self.physical_groups = physical_groups

        # properties calculated internally by the class
        self.solid = None
        self.wire = None
        self.render_mesh = None

        self.processed_points = None
        # self.volume = None
        self.hash_value = None
        self.points_hash_value = None
        self.x_min = None
        self.x_max = None
        self.z_min = None
        self.z_max = None
        self.graveyard_offset = None  # set by the make_graveyard method
        self.patch = None

    @property
    def graveyard_size(self):
        return self._graveyard_size

    @graveyard_size.setter
    def graveyard_size(self, value):
        if value is None:
            self._graveyard_size = None
        elif not isinstance(value, (float, int)):
            raise TypeError("graveyard_size must be a number")
        elif value < 0:
            raise ValueError("graveyard_size must be positive")
        self._graveyard_size = value

    @property
    def graveyard_offset(self):
        return self._graveyard_offset

    @graveyard_offset.setter
    def graveyard_offset(self, value):
        if value is None:
            self._graveyard_offset = None
        elif not isinstance(value, (float, int)):
            raise TypeError("graveyard_offset must be a number")
        elif value < 0:
            raise ValueError("graveyard_offset must be positive")
        self._graveyard_offset = value

    @property
    def solid(self):
        """The CadQuery solid of the 3d object. Returns a CadQuery workplane
        or CadQuery Compound"""

        ignored_keys = ["_solid", "_hash_value"]
        if get_hash(self, ignored_keys) != self.hash_value:
            self.create_solid()
            self.hash_value = get_hash(self, ignored_keys)

        return self._solid

    @solid.setter
    def solid(self, value):
        self._solid = value

    @property
    def wire(self):
        """The CadQuery wire of the 3d object. Returns a CadQuery workplane
        or CadQuery Compound"""

        ignored_keys = ["_wire", "_solid", "_hash_value"]
        if get_hash(self, ignored_keys) != self.hash_value:
            self.create_solid()
            self.hash_value = get_hash(self, ignored_keys)

        return self._wire

    @wire.setter
    def wire(self, value):
        self._wire = value

    @property
    def cut(self):
        return self._cut

    @cut.setter
    def cut(self, value):
        self._cut = value

    @property
    def intersect(self):
        return self._intersect

    @intersect.setter
    def intersect(self, value):
        self._intersect = value

    @property
    def union(self):
        return self._union

    @union.setter
    def union(self, value):
        self._union = value

    @property
    def largest_dimension(self):
        """Calculates a bounding box for the Shape and returns the largest
        absolute value of the largest dimension of the bounding box"""
        largest_dimension = 0
        if isinstance(self.solid, (Compound, shapes.Solid)):
            for solid in self.solid.Solids():
                bound_box = solid.BoundingBox()
                largest_dimension = max(
                    abs(bound_box.xmax),
                    abs(bound_box.xmin),
                    abs(bound_box.ymax),
                    abs(bound_box.ymin),
                    abs(bound_box.zmax),
                    abs(bound_box.zmin),
                    largest_dimension
                )
        else:
            bound_box = self.solid.val().BoundingBox()
            largest_dimension = max(
                abs(bound_box.xmax),
                abs(bound_box.xmin),
                abs(bound_box.ymax),
                abs(bound_box.ymin),
                abs(bound_box.zmax),
                abs(bound_box.zmin),
            )
        self.largest_dimension = largest_dimension
        return largest_dimension

    @largest_dimension.setter
    def largest_dimension(self, value):
        self._largest_dimension = value

    @property
    def workplane(self):
        return self._workplane

    @workplane.setter
    def workplane(self, value):
        if isinstance(value, Plane):
            self._workplane = value
        elif isinstance(value, str):
            acceptable_values = ["XY", "YZ", "XZ", "YX", "ZY", "ZX"]
            if value in acceptable_values:
                self._workplane = value
            else:
                raise ValueError(
                    "Shape.workplane must be one of ",
                    acceptable_values,
                    " not ",
                    value)
        else:
            raise TypeError(
                "Shape.workplane must be a string or a ",
                "cadquery.Plane object")

    @property
    def rotation_axis(self):
        return self._rotation_axis

    @rotation_axis.setter
    def rotation_axis(self, value):
        if isinstance(value, str):
            acceptable_values = \
                ["X", "Y", "Z", "-X", "-Y", "-Z", "+X", "+Y", "+Z"]
            if value not in acceptable_values:
                msg = "Shape.rotation_axis must be one of " + \
                    " ".join(acceptable_values) + \
                    " not " + value
                raise ValueError(msg)
        elif isinstance(value, Iterable):
            msg = "Shape.rotation_axis must be a tuple of three floats (X, Y, Z)"
            if len(value) != 2:
                raise ValueError(msg)
            for point in value:
                if not isinstance(point, tuple):
                    raise ValueError(msg)
                if len(point) != 3:
                    raise ValueError(msg)
                for val in point:
                    if not isinstance(val, (int, float)):
                        raise ValueError(msg)

            if value[0] == value[1]:
                msg = "The two points must be different"
                raise ValueError(msg)
        elif value is not None:
            msg = "Shape.rotation_axis must be a list or a string or None"
            raise ValueError(msg)
        self._rotation_axis = value

    @property
    def area(self):
        """Get the total surface area of the Shape. Returns a float"""
        if isinstance(self.solid, Compound):
            return self.solid.Area()

        return self.solid.val().Area()

    @property
    def areas(self):
        """Get the surface areas of the Shape. Compound shapes provide a
        separate area value for each entry. Returns a list of floats"""
        all_areas = []
        if isinstance(self.solid, Compound):
            for face in self.solid.Faces():
                all_areas.append(face.Area())
            return all_areas

        for face in self.solid.val().Faces():
            all_areas.append(face.Area())
        return all_areas

    @property
    def hash_value(self):
        return self._hash_value

    @hash_value.setter
    def hash_value(self, value):
        self._hash_value = value

    @property
    def points_hash_value(self):
        return self._points_hash_value

    @points_hash_value.setter
    def points_hash_value(self, value):
        self._points_hash_value = value

    @property
    def color(self):
        return self._color

    @color.setter
    def color(self, value):
        if isinstance(value, (list, tuple)):
            if len(value) in [3, 4]:
                for i in value:
                    if not isinstance(i, (int, float)):
                        raise ValueError(
                            "Individual entries in the Shape.color must a "
                            "number (float or int)")
                    if i > 1 or i < 0:
                        raise ValueError(
                            "Individual entries in the Shape.color must be "
                            "between 0 and 1"
                        )
            else:
                raise ValueError(
                    "Shape.color must be a list or tuple of 3 or 4 floats")
        else:
            raise ValueError(
                "Shape.color must be a list or tuple")

        self._color = value

    @property
    def tet_mesh(self):
        return self._tet_mesh

    @tet_mesh.setter
    def tet_mesh(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError("Shape.tet_mesh must be a string", value)
        self._tet_mesh = value

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        if value is not None and not isinstance(value, float):
            raise ValueError("Shape.scale must be a float", value)
        self._scale = value

    @property
    def name(self):
        """The name of the Shape, used to identify Shapes when exporting_html
        """
        return self._name

    @name.setter
    def name(self, value):
        if value is not None and not isinstance(value, str):
            raise ValueError("Shape.name must be a string", value)
        self._name = value

    @property
    def processed_points(self):
        """Shape.processed_points attributes is set internally from the
        Shape.points"""

        if self.points is not None:
            # if .points have changed since last time this was run
            if self.old_points != self.points:
                # assign current .points value to .old_points
                self.old_points = self.points

                # compute .processed_points
                if self.connection_type == "mixed":
                    values = self.points
                else:
                    values = [(*p, self.connection_type) for p in self.points]

                if values[0][:2] != values[-1][:2]:
                    values.append(values[0])

                self._processed_points = values
            return self._processed_points
        return None

    @processed_points.setter
    def processed_points(self, value):
        self._processed_points = value

    @property
    def points(self):
        """Sets the Shape.point attributes.

        Args:
            points (a tuple of tuples): tuple of points that create the
                shape

        Raises:
            incorrect type: only list of lists or tuples are accepted
        """
        ignored_keys = ["_points", "_points_hash_value"]
        if hasattr(self, 'find_points') and \
                self.points_hash_value != get_hash(self, ignored_keys):
            self.find_points()
            self.points_hash_value = get_hash(self, ignored_keys)

        return self._points

    @points.setter
    def points(self, values_in):
        if values_in is None:
            values = values_in
        else:
            values = values_in[:]
        if values is not None:
            if not isinstance(values, (list, tuple)):
                raise ValueError("points must be a list or a tuple")

            for value in values:
                if not isinstance(value, (list, tuple)):
                    msg = (f'individual points must be a tuple.{value} in of '
                           f'type {type(value)}')
                    raise ValueError(msg)

            for counter, value in enumerate(values):
                if self.connection_type == 'mixed':
                    if len(value) != 3:
                        if counter != len(
                                values) - 1:  # last point doesn't need connections
                            msg = (
                                'individual points should contain 3 '
                                'entries when the Shape.connection_type is '
                                '"mixed". The entries should contain two '
                                f'coordinates and a connection type. {value} '
                                'has a length of {len(value)}')
                            print(values)
                            raise ValueError(msg)
                else:
                    if len(value) != 2:
                        msg = ('individual points should contain 2 entries '
                               'when the Shape.connection_type is '
                               f'{self.connection_type}. The entries should '
                               f'just contain the coordinates {value} has a '
                               'length of {len(value)}')
                        raise ValueError(msg)

                # Checks that the XY points are numbers
                if not isinstance(value[0], numbers.Number):
                    msg = (
                        'The first value in the tuples that make up the '
                        'points represents the X value and must be a number '
                        f'{value}')
                    raise ValueError(msg)
                if not isinstance(value[1], numbers.Number):
                    msg = ('The second value in the tuples that make up the '
                           'points represents the X value and must be a '
                           f'number {value}')
                    raise ValueError(msg)

                # Checks that only straight and spline are in the connections
                # part of points
                if len(value) == 3:
                    if value[2] not in ["straight", "spline", "circle"]:
                        msg = ('individual connections must be either '
                               '"straight", "circle" or "spline"')
                        raise ValueError(msg)

            if len(values) > 1:
                if values[0][:2] == values[-1][:2]:
                    msg = ('The coordinates of the last and first points are '
                           'the same.')
                    raise ValueError(msg)

        self._points = values

    @property
    def azimuth_placement_angle(self):
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        error = False
        if isinstance(value, (int, float, Iterable)) and \
                not isinstance(value, str):
            if isinstance(value, Iterable):
                for i in value:
                    if not isinstance(i, (int, float)):
                        error = True
        else:
            error = True

        if error:
            msg = "azimuth_placement_angle must be a float or list of floats"
            raise ValueError(msg)
        self._azimuth_placement_angle = value

    def from_stp_file(self, filename: str):
        """Loads the filename using CadQuery and populates the Shape.solid
        with the contents

        Args:
            filename: the file name of the stp / step file to be loaded
        """
        result = importers.importStep(filename)
        self.solid = result

    def show(self, default_edgecolor: Tuple[float, float, float] = (0, 0, 0)):
        """Shows / renders the CadQuery the 3d object in Jupyter Lab. Imports
        show from jupyter_cadquery.cadquery and returns show(Shape.solid)

        Args:
            default_edgecolor: the color to use for the edges, passed to
                jupyter_cadquery.cadquery show. Tuple of three values expected
                individual values in the tuple should be floats between 0. and
                1.

        Returns:
            jupyter_cadquery.cadquery.show object
        """

        try:
            from jupyter_cadquery.cadquery import Part, PartGroup, show
        except ImportError:
            msg = (
                'To use Shape.show() you must install jupyter_cadquery. To'
                'install jupyter_cadquery type "pip install jupyter_cadquery"'
                ' in the terminal')
            raise ImportError(msg)

        parts = []
        if self.name is None:
            name = 'Shape.name not set'
        else:
            name = self.name

        scaled_color = [int(i * 255) for i in self.color[0:3]]
        scaled_edge_color = [int(i * 255) for i in default_edgecolor[0:3]]
        if isinstance(
                self.solid,
                (shapes.Shape, shapes.Compound)):
            for i, solid in enumerate(self.solid.Solids()):
                parts.append(
                    Part(
                        solid,
                        name=f"{name}{i}",
                        color=scaled_color,
                        show_edges=True
                    ))
        else:
            parts.append(
                Part(
                    self.solid.val(),
                    name=f"{name}",
                    color=scaled_color,
                    show_edges=True
                ))

        return show(PartGroup(parts), default_edgecolor=scaled_edge_color)

    def create_solid(self) -> Workplane:
        solid = None
        if self.processed_points is not None:
            # obtains the first two values of the points list
            XZ_points = [(p[0], p[1]) for p in self.processed_points]

            # obtains the last values of the points list
            connections = [p[2] for p in self.processed_points[:-1]]

            current_linetype = connections[0]
            current_points_list = []
            instructions = []
            # groups together common connection types
            for i, connection in enumerate(connections):
                if connection == current_linetype:
                    current_points_list.append(XZ_points[i])
                else:
                    current_points_list.append(XZ_points[i])
                    instructions.append(
                        {current_linetype: current_points_list})
                    current_linetype = connection
                    current_points_list = [XZ_points[i]]
            instructions.append({current_linetype: current_points_list})

            if list(instructions[-1].values())[0][-1] != XZ_points[0]:
                keyname = list(instructions[-1].keys())[0]
                instructions[-1][keyname].append(XZ_points[0])

            if hasattr(self, "path_points"):

                factor = 1
                if self.workplane in ["XZ", "YX", "ZY"]:
                    factor *= -1

                solid = Workplane(self.workplane).center(0, 0)

                if self.force_cross_section:
                    for point in self.path_points[:-1]:
                        solid = solid.workplane(offset=point[1] * factor).\
                            center(point[0], 0).workplane()
                        for entry in instructions:
                            connection_type = list(entry.keys())[0]
                            if connection_type == "spline":
                                solid = solid.spline(
                                    listOfXYTuple=list(entry.values())[0])
                            elif connection_type == "straight":
                                solid = solid.polyline(list(entry.values())[0])
                            elif connection_type == "circle":
                                p0, p1, p2 = list(entry.values())[0][:3]
                                solid = solid.moveTo(p0[0], p0[1]).\
                                    threePointArc(p1, p2)
                        solid = solid.close()
                        solid = solid.center(-point[0], 0).\
                            workplane(offset=-point[1] * factor)

                elif self.force_cross_section is False:
                    solid = solid.workplane(
                        offset=self.path_points[0][1] *
                        factor).center(
                        self.path_points[0][0],
                        0).workplane()
                    for entry in instructions:
                        connection_type = list(entry.keys())[0]
                        if connection_type == "spline":
                            solid = solid.spline(
                                listOfXYTuple=list(entry.values())[0])
                        elif connection_type == "straight":
                            solid = solid.polyline(list(entry.values())[0])
                        elif connection_type == "circle":
                            p0 = list(entry.values())[0][0]
                            p1 = list(entry.values())[0][1]
                            p2 = list(entry.values())[0][2]
                            solid = solid.moveTo(
                                p0[0], p0[1]).threePointArc(
                                p1, p2)

                    solid = solid.close().center(0, 0).\
                        center(-self.path_points[0][0], 0).\
                        workplane(offset=-self.path_points[0][1] * factor)

                solid = solid.workplane(offset=self.path_points[-1][1] * factor).\
                    center(self.path_points[-1][0], 0).workplane()

            else:
                # for rotate and extrude shapes
                solid = Workplane(self.workplane)
                # for extrude shapes
                if hasattr(self, "extrusion_start_offset"):
                    extrusion_offset = -self.extrusion_start_offset
                    solid = solid.workplane(offset=extrusion_offset)

            for entry in instructions:
                if list(entry.keys())[0] == "spline":
                    solid = solid.spline(listOfXYTuple=list(entry.values())[0])
                if list(entry.keys())[0] == "straight":
                    solid = solid.polyline(list(entry.values())[0])
                if list(entry.keys())[0] == "circle":
                    p0 = list(entry.values())[0][0]
                    p1 = list(entry.values())[0][1]
                    p2 = list(entry.values())[0][2]
                    solid = solid.moveTo(p0[0], p0[1]).threePointArc(p1, p2)

        return solid

    def rotate_solid(
            self,
            solid: Optional[Workplane]) -> Workplane:
        # Checks if the azimuth_placement_angle is a list of angles
        if isinstance(self.azimuth_placement_angle, Iterable):
            azimuth_placement_angles = self.azimuth_placement_angle
        else:
            azimuth_placement_angles = [self.azimuth_placement_angle]

        rotated_solids = []
        # Perform seperate rotations for each angle
        for angle in azimuth_placement_angles:
            rotated_solids.append(
                solid.rotate(
                    *self.get_rotation_axis()[0], angle))
        solid = Workplane(self.workplane)

        # Joins the seperate solids together
        for i in rotated_solids:
            solid = solid.union(i)
        return solid

    def get_rotation_axis(self):
        # TODO add return type hinting -> Tuple[List[Tuple[int, int, int],
        # Tuple[int, int, int]], str]
        """Returns the rotation axis for a given shape. If self.rotation_axis
        is None, the rotation axis will be computed from self.workplane (or
        from self.path_workplane if applicable). If self.rotation_axis is an
        acceptable string (eg. "X", "+Y", "-Z"...) then this axis will be used.
        If self.rotation_axis is a list of two points, then these two points
        will be used to form an axis.

        Returns:
            list, str: list of two XYZ points and the string of the axis (eg.
                "X", "Y"..)
        """
        rotation_axis = {
            "X": [(-1, 0, 0), (1, 0, 0)],
            "-X": [(1, 0, 0), (-1, 0, 0)],
            "Y": [(0, -1, 0), (0, 1, 0)],
            "-Y": [(0, 1, 0), (0, -1, 0)],
            "Z": [(0, 0, -1), (0, 0, 1)],
            "-Z": [(0, 0, 1), (0, 0, -1)],
        }
        if isinstance(self.rotation_axis, str):
            # X, Y or Z axis
            return (
                rotation_axis[self.rotation_axis.replace("+", "")],
                self.rotation_axis
            )
        elif isinstance(self.rotation_axis, Iterable):
            # Custom axis
            return self.rotation_axis, "custom_axis"
        elif self.rotation_axis is None:
            # Axis from workplane or path_workplane
            if hasattr(self, "path_workplane"):
                # compute from path_workplane instead
                workplane = self.path_workplane
            else:
                workplane = self.workplane
            return rotation_axis[workplane[1]], workplane[1]

    def create_limits(self) -> Tuple[float, float, float, float]:
        """Finds the x,y,z limits (min and max) of the points that make up the
        face of the shape. Note the Shape may extend beyond this boundary if
        splines are used to connect points.

        Raises:
            ValueError: if no points are defined

        Returns:
            float, float, float, float, float, float: x_minimum, x_maximum,
            y_minimum, y_maximum, z_minimum, z_maximum
        """

        if hasattr(self, "find_points"):
            self.find_points()
        if self.points is None:
            raise ValueError("No points defined for", self)

        self.x_min = float(min([row[0] for row in self.points]))
        self.x_max = float(max([row[0] for row in self.points]))

        self.z_min = float(min([row[1] for row in self.points]))
        self.z_max = float(max([row[1] for row in self.points]))

        return self.x_min, self.x_max, self.z_min, self.z_max

    def export_stl(
            self,
            filename: str,
            tolerance: Optional[float] = 0.001,
            angular_tolerance: Optional[float] = 0.1,
            verbose: Optional[bool] = True) -> str:
        """Exports an stl file for the Shape.solid.

        Args:
            filename: the filename of exported the stl file. Defaults to None
                which will attempt to use the Shape.stl_filename. If both are
                None then a valueError will be raised.
            tolerance: the deflection tolerance of the faceting
            angular_tolerance: the angular tolerance, in radians
            verbose: Enables (True) or disables (False) the printing of the
                file produced.
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".stl":
            msg = f'filename should end with .stl, not {path_filename.suffix}'
            raise ValueError(msg)

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        exporters.export(self.solid, str(path_filename), exportType='STL',
                         tolerance=tolerance,
                         angularTolerance=angular_tolerance)

        if verbose:
            print("Saved file as ", path_filename)

        return str(path_filename)

    def export_brep(
        self,
        filename
    ):
        """Exports a brep file for the Shape.solid.

        Args:
            filename: the filename of exported the brep file.
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".brep":
            msg = "When exporting a brep file the filename must end with .brep"
            raise ValueError(msg)

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        self.solid.val().exportBrep(str(path_filename))
        # alternative method is to use BRepTools that might support imprinting
        # and merging https://github.com/CadQuery/cadquery/issues/449
        # from OCP.BRepTools import BRepTools
        # BRepTools.Write_s(self.solid.toOCC(), str(path_filename))

        return str(path_filename)

    def export_stp(
            self,
            filename: str,
            units: Optional[str] = 'mm',
            mode: Optional[str] = 'solid',
            verbose: Optional[bool] = True) -> str:
        """Exports an stp file for the Shape.solid.

        Args:
            filename: the filename of exported the stp file.
            units: the units of the stp file, options are 'cm' or 'mm'.
                Default is mm.
            mode: the object to export can be either
                'solid' which exports 3D solid shapes or the 'wire' which
                exports the wire edges of the shape. Defaults to 'solid'.
            verbose: Enables (True) or disables (False) the printing of the
                file produced.
        """

        path_filename = Path(filename)

        if path_filename.suffix == ".stp" or path_filename.suffix == ".step":
            pass
        else:
            msg = f'filename should end with .stp or .step, not {path_filename.suffix}'
            raise ValueError(msg)

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        if mode == 'solid':

            assembly = Assembly(name=self.name)

            if self.color is None:
                assembly.add(self.solid)
            else:
                assembly.add(self.solid, color=Color(*self.color))

            assembly.save(str(path_filename), exportType='STEP')

            # previous method does not support colours but puts the solid in the base file level
            # exporters.export(self.solid, str(path_filename), exportType='STEP')

        elif mode == 'wire':
            exporters.export(self.wire, str(path_filename), exportType='STEP')
        else:
            raise ValueError("The mode argument for export_stp \
                only accepts 'solid' or 'wire'", self)

        if units == 'cm':
            _replace(
                path_filename,
                'SI_UNIT(.MILLI.,.METRE.)',
                'SI_UNIT(.CENTI.,.METRE.)')

        if verbose:
            print(f"Saved file as {path_filename}")

        return str(path_filename)

    def export_svg(
            self,
            filename: Optional[str] = 'shape.svg',
            projectionDir: Tuple[float, float, float] = (-1.75, 1.1, 5),
            width: Optional[float] = 800,
            height: Optional[float] = 800,
            marginLeft: Optional[float] = 100,
            marginTop: Optional[float] = 100,
            strokeWidth: Optional[float] = None,
            strokeColor: Optional[Tuple[int, int, int]] = (0, 0, 0),
            hiddenColor: Optional[Tuple[int, int, int]] = (100, 100, 100),
            showHidden: Optional[bool] = True,
            showAxes: Optional[bool] = False) -> str:
        """Exports an svg file for the Reactor.solid. If the filename provided
        doesn't end with .svg it will be added.

        Args:
            filename: the filename of the svg file to be exported. Defaults to
                "reactor.svg".
            projectionDir: The direction vector to view the geometry from
                (x, y, z). Defaults to (-1.75, 1.1, 5)
            width: the width of the svg image produced in pixels. Defaults to
                1000
            height: the height of the svg image produced in pixels. Defaults to
                800
            marginLeft: the number of pixels between the left edge of the image
                and the start of the geometry.
            marginTop: the number of pixels between the top edge of the image
                and the start of the geometry.
            strokeWidth: the width of the lines used to draw the geometry.
                Defaults to None which automatically selects an suitable width.
            strokeColor: the color of the lines used to draw the geometry in
                RGB format with each value between 0 and 255. Defaults to
                (0, 0, 0) which is black.
            hiddenColor: the color of the lines used to draw the geometry in
                RGB format with each value between 0 and 255. Defaults to
                (100, 100, 100) which is light grey.
            showHidden: If the edges obscured by geometry should be included in
                the diagram. Defaults to True.
            showAxes: If the x, y, z axis should be included in the image.
                Defaults to False.

        Returns:
            str: the svg filename created
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".svg":
            path_filename = path_filename.with_suffix(".svg")

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        opt = {
            "width": width,
            "height": height,
            "marginLeft": marginLeft,
            "marginTop": marginTop,
            "showAxes": showAxes,
            "projectionDir": projectionDir,
            "strokeColor": strokeColor,
            "hiddenColor": hiddenColor,
            "showHidden": showHidden
        }

        if strokeWidth is not None:
            opt["strokeWidth"] = strokeWidth

        exporters.export(self.solid, str(path_filename), exportType='SVG',
                         opt=opt)

        print("Saved file as ", path_filename)

        return str(path_filename)

    def export_html_3d(
            self,
            filename: Optional[str] = "shape_3d.html",
    ):
        """Saves an interactive 3d html view of the Shape to a html file.

        Args:
            filename: the filename used to save the html graph. Defaults to
                shape_3d.html

        Returns:
            str: filename of the created html file
        """

        from ipywidgets.embed import embed_minimal_html

        view = self.show()

        if view is None:
            return None

        embed_minimal_html(
            filename,
            views=[view.cq_view.renderer],
            title='Renderer'
        )

        return filename

    def export_html(
            self,
            filename: Optional[str] = "shape.html",
            facet_splines: Optional[bool] = True,
            facet_circles: Optional[bool] = True,
            tolerance: Optional[float] = 1e-3,
            view_plane: Optional[str] = None,
    ):
        """Creates a html graph representation of the points and connections
        for the Shape object. Shapes are colored by their .color property.
        Shapes are also labelled by their .name. If filename provided doesn't
        end with .html then .html will be added.

        Args:
            filename: the filename used to save the html graph. Defaults to
                shape.html
            facet_splines: If True then spline edges will be faceted. Defaults
                to True.
            facet_circles: If True then circle edges will be faceted. Defaults
                to True.
            tolerance: faceting toleranceto use when faceting cirles and
                splines. Defaults to 1e-3.
            view_plane: The plane to project. Options are 'XZ', 'XY', 'YZ',
                'YX', 'ZY', 'ZX', 'RZ' and 'XYZ'. Defaults to 'RZ'. Defaults to
                the workplane of the paramak.Shape.

        Returns:
            plotly.Figure(): figure object
        """

        # if view plane is not set then use the shape workplane
        if view_plane is None:
            view_plane = self.workplane

        if self.solid is None:
            raise ValueError("No solid was found for ", self)

        if isinstance(self.solid, Workplane):
            edges = self.solid.val().Edges()
        else:
            edges = self.solid.Edges()

        fig = paramak.utils.export_wire_to_html(
            wires=edges,
            filename=None,
            view_plane=view_plane,
            facet_splines=facet_splines,
            facet_circles=facet_circles,
            tolerance=tolerance,
            title=(f'coordinates of {self.__class__.__name__} shape, viewed '
                   'from the {view_plane} plane')
        )

        if self.points is not None:
            fig.add_trace(
                plotly_trace(
                    points=self.points,
                    mode="markers",
                    name='Shape.points'
                )
            )

        # sweep shapes have .path_points but not .points attribute
        if hasattr(self, 'path_points'):
            fig.add_trace(
                plotly_trace(
                    points=self.path_points,
                    mode="markers",
                    name='Shape.path_points'
                )
            )

        if filename is not None:

            Path(filename).parents[0].mkdir(parents=True, exist_ok=True)

            path_filename = Path(filename)

            if path_filename.suffix != ".html":
                path_filename = path_filename.with_suffix(".html")

            fig.write_html(str(path_filename))

        return fig

    def export_2d_image(
            self,
            filename: Optional[str] = 'shape.png',
            xmin: Optional[float] = 0.,
            xmax: Optional[float] = 900.,
            ymin: Optional[float] = -600.,
            ymax: Optional[float] = 600.):
        """Exports a 2d image (png) of the reactor. Components are colored by
        their Shape.color property. If filename provided doesn't end with .png
        then .png will be added.

        Args:
            filename: the filename of the saved png image.
            xmin: the minimum x value of the x axis.
                Defaults to 0..
            xmax: the maximum x value of the x axis.
                Defaults to 900..
            ymin: the minimum y value of the y axis.
                Defaults to -600..
            ymax: the maximum y value of the y axis.
                Defaults to 600..

        Returns:
            matplotlib.plt(): a plt object
        """

        fig, ax = plt.subplots()

        patch = self._create_patch()

        ax.add_collection(patch)

        ax.axis("equal")
        ax.set(xlim=(xmin, xmax), ylim=(ymin, ymax))
        ax.set_aspect("equal", "box")

        plt.savefig(filename, dpi=100)
        plt.close()
        print(f'\n saved 2d image to {filename}', )

        return plt

    def _create_patch(self):
        """Creates a matplotlib polygon patch from the Shape points. This is
        used when making 2d images of the Shape object.

        Raises:
            ValueError: No points defined for the Shape

        Returns:
            Matplotlib object patch: a plotable polygon shape
        """

        if self.processed_points is None:
            raise ValueError("No processed_points defined for", self)

        patches = []

        edges = facet_wire(
            wire=self.wire,
            facet_splines=True,
            facet_circles=True)

        fpoints = []
        for edge in edges:
            for vertice in edge.Vertices():
                fpoints.append((vertice.X, vertice.Z))

        polygon = Polygon(fpoints, closed=True)
        patches.append(polygon)

        patch = PatchCollection(patches)

        if self.color is not None:
            print('color is ', self.color)
            patch.set_facecolor(self.color[0:3])
            patch.set_color(self.color[0:3])
            patch.color = self.color[0:3]
            patch.edgecolor = self.color[0:3]
            # checks to see if an alpha value is provided in the color
            if len(self.color) == 4:
                patch.set_alpha = self.color[-1]
        self.patch = patch
        return patch

    def perform_boolean_operations(self, solid: Workplane, **kwargs):
        """Performs boolean cut, intersect and union operations if shapes are
        provided"""

        # If a cut solid is provided then perform a boolean cut
        if self.cut is not None:
            solid = cut_solid(solid, self.cut)

        # If a wedge cut is provided then perform a boolean cut
        # Performed independently to avoid use of self.cut
        # Prevents repetition of 'outdated' wedge cuts
        if 'wedge_cut' in kwargs:
            if kwargs['wedge_cut'] is not None:
                solid = cut_solid(solid, kwargs['wedge_cut'])

        # If an intersect is provided then perform a boolean intersect
        if self.intersect is not None:
            solid = intersect_solid(solid, self.intersect)

        # If an intersect is provided then perform a boolean intersect
        if self.union is not None:
            solid = union_solid(solid, self.union)

        return solid

    def make_graveyard(
            self,
            graveyard_size: Optional[float] = None,
            graveyard_offset: Optional[float] = None,
    ):
        """Creates a graveyard volume (bounding box) that encapsulates all
        volumes. This is required by DAGMC when performing neutronics
        simulations. The graveyard size can be ascertained in two ways. Either
        the size can be set directly using the graveyard_size which is the
        quickest method. Alternativley the graveyard can be automatically sized
        to the geometry by setting a graveyard_offset value. If both options
        are set then the method will default to using the graveyard_size
        preferentially.

        Args:
            graveyard_size: directly sets the size of the graveyard. Defaults
                to None which then uses the Reactor.graveyard_size attribute.
            graveyard_offset: the offset between the largest edge of the
                geometry and inner bounding shell created. Defaults to None
                which then uses Reactor.graveyard_offset attribute.

        Returns:
            paramak.HollowCube: a shell volume that bounds the geometry,
                referred to as a graveyard in DAGMC.
        """

        if graveyard_size is not None:
            graveyard_size_to_use = graveyard_size

        elif self.graveyard_size is not None:
            graveyard_size_to_use = self.graveyard_size

        elif graveyard_offset is not None:
            self.solid
            graveyard_size_to_use = self.largest_dimension * 2 + graveyard_offset * 2

        elif self.graveyard_offset is not None:
            self.solid
            graveyard_size_to_use = self.largest_dimension * 2 + self.graveyard_offset * 2

        else:
            raise ValueError(
                "the graveyard_size, Shape.graveyard_size, "
                "graveyard_offset and Shape.graveyard_offset are all None. "
                "Please specify at least one of these attributes or arguments")

        graveyard_shape = paramak.HollowCube(
            length=graveyard_size_to_use,
            name="graveyard",
        )

        self.graveyard = graveyard_shape

        return graveyard_shape

    def export_graveyard(
            self,
            filename: Optional[str] = "graveyard.stp",
            graveyard_offset: Optional[float] = 100) -> str:
        """Writes an stp file (CAD geometry) for the reactor graveyard. This
        is needed for DAGMC simulations. This method also calls
        Reactor.make_graveyard with the offset.

        Args:
            filename (str): the filename for saving the stp file
            graveyard_offset (float): the offset between the largest edge of
                the geometry and inner bounding shell created. Defaults to
                Reactor.graveyard_offset

        Returns:
            str: the stp filename created
        """

        self.make_graveyard(graveyard_offset=graveyard_offset)
        new_filename = self.graveyard.export_stp(Path(filename))

        return new_filename

    def convert_all_circle_connections_to_splines(
            self,
            tolerance: Optional[float] = 0.1
    ) -> List[Tuple[float, float, str]]:
        """Replaces circle edges in Shape.processed_points points with spline
        edges. The spline control coordinates are obtained by faceting the
        circle edge with the provided tolerance. The Shape.processed_points
        will be updated to exclude the circle points and include the new spline
        points. This method works best when the connection before and after the
        circle is a straight connection type. This method is useful when
        converting the stp file into other formats due to errors in the
        conversion of circle edges.

        Args:
            tolerance: the precision of the faceting.

        Returns:
            The new points with spline connections
        """

        new_points = []
        counter = 0
        while counter < len(self.processed_points):

            if self.processed_points[counter][2] == 'circle':
                p_0 = self.processed_points[counter][:2]
                p_1 = self.processed_points[counter + 1][:2]
                p_2 = self.processed_points[counter + 2][:2]

                points = paramak.utils.convert_circle_to_spline(
                    p_0, p_1, p_2, tolerance=tolerance
                )

                # the last point needs to have the connection type of p2
                for point in points[:-1]:
                    new_points.append((point[0], point[1], 'spline'))

                new_points.append(self.processed_points[counter + 2])
                counter = counter + 3
            else:
                new_points.append(self.processed_points[counter])
                counter = counter + 1

        # @jon I'm not 100% if this change is correct or not
        self.processed_points = new_points
        return new_points

    def volume(self, split_compounds=False) -> Union[float, List[float]]:
        """Get the total volume of the Shape.

        Args:
            split_compounds: If the Shape is a compound of Shapes and therefore
                contains multiple volumes. This option allows access to the
                separate volumes of each component within a Shape (True) or the
                volumes of compounds can be summed (False).

        Returns:
            The the volume(s) of the Shape
        """

        if not isinstance(split_compounds, bool):
            msg = f'split_compounds must be True or False. Not {split_compounds}'
            raise ValueError(msg)

        # returns a list of floats
        if split_compounds:
            all_volumes = []
            if isinstance(self.solid, Compound):
                for solid in self.solid.Solids():
                    all_volumes.append(solid.Volume())
                return all_volumes

            return [self.solid.val().Volume()]

        # returns a float

        if isinstance(self.solid, Compound):
            return self.solid.Volume()

        return self.solid.val().Volume()




import unittest

import paramak
import pytest


class TestCoolantChannelRingCurved(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.CoolantChannelRingCurved(
            height=100,
            channel_radius=10,
            ring_radius=70,
            mid_offset=-20,
            number_of_coolant_channels=6
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a CoolantChannelRingCurved are correct."""

        # assert self.test_shape.rotation_angle == 360
        assert self.test_shape.start_angle == 0

    def test_creation(self):
        """Creates a coolant channel ring using the CoolantChannelRingCurved parametric shape
        and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_relative_volumes(self):
        """Creates coolant channel rings using the CoolantChannelRingCurved parametric shape
        and checks the relative volumes are correct."""

        test_volume = self.test_shape.volume()
        self.test_shape.number_of_coolant_channels = 3
        assert test_volume == pytest.approx(self.test_shape.volume() * 2)

        test_volume = self.test_shape.volume()
        self.test_shape.mid_offset = -30
        assert test_volume > self.test_shape.volume()
        self.test_shape.force_cross_section = True
        assert test_volume < self.test_shape.volume()

    def test_start_angle(self):
        """Checks that the coolant channels are placed at the correct azimuthal placement
        angles for a given start angle."""

        assert self.test_shape.azimuth_placement_angle == [
            0, 60, 120, 180, 240, 300
        ]
        self.test_shape.start_angle = 10
        assert self.test_shape.azimuth_placement_angle == [
            10, 70, 130, 190, 250, 310
        ]




import cadquery as cq
from paramak import RotateStraightShape


class PoloidalFieldCoilSet(RotateStraightShape):
    """Creates a series of rectangular poloidal field coils.

    Args:
        heights (list of floats): the vertical (z axis) heights of the coils
            (cm).
        widths (list of floats): the horizontal (x axis) widths of the coils
            (cm).
        center_points (list of tuple of floats): the center of the coil (x,z)
            values e.g. [(100,100), (100,200)] (cm).
        name (str, optional): defaults to "pf_coil".
    """

    def __init__(
        self,
        heights,
        widths,
        center_points,
        name="pf_coil",
        **kwargs
    ):

        super().__init__(
            name=name,
            **kwargs
        )

        self.center_points = center_points
        self.heights = heights
        self.widths = widths

        if len(
            self.widths) != len(
            self.heights) or len(
            self.heights) != len(
                self.center_points):
            raise ValueError("The length of widthts, height and center_points \
                must be the same when making a PoloidalFieldCoilSet")

    @property
    def center_points(self):
        return self._center_points

    @center_points.setter
    def center_points(self, value):
        if not isinstance(value, list):
            raise ValueError(
                "PoloidalFieldCoilSet center_points must be a list")
        self._center_points = value

    @property
    def heights(self):
        return self._heights

    @heights.setter
    def heights(self, value):
        if not isinstance(value, list):
            raise ValueError("PoloidalFieldCoilSet heights must be a list")
        self._heights = value

    @property
    def widths(self):
        return self._widths

    @widths.setter
    def widths(self, value):
        if not isinstance(value, list):
            raise ValueError("PoloidalFieldCoilSet widths must be a list")
        self._widths = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the poloidal field coil shape."""

        all_points = []

        for height, width, center_point in zip(
                self.heights, self.widths, self.center_points):

            all_points = all_points + [
                (
                    center_point[0] + width / 2.0,
                    center_point[1] + height / 2.0,
                ),  # upper right
                (
                    center_point[0] + width / 2.0,
                    center_point[1] - height / 2.0,
                ),  # lower right
                (
                    center_point[0] - width / 2.0,
                    center_point[1] - height / 2.0,
                ),  # lower left
                (
                    center_point[0] - width / 2.0,
                    center_point[1] + height / 2.0,
                ),  # upper left
            ]

        self.points = all_points

    def create_solid(self):
        """Creates a 3d solid using points with straight connections
        edges, azimuth_placement_angle and rotation angle.
        individual solids in the compound can be accessed using .Solids()[i]
        where i is an int

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        iter_points = iter(self.points)
        pf_coils_set = []
        wires = []
        for p1, p2, p3, p4 in zip(
                iter_points, iter_points, iter_points, iter_points):

            solid = (
                cq.Workplane(self.workplane)
                .polyline([p1[:2], p2[:2], p3[:2], p4[:2]])
            )

            wire = solid.close()

            wires.append(wire)

            solid = wire.revolve(self.rotation_angle)

            pf_coils_set.append(solid)

        compound = cq.Compound.makeCompound(
            [a.val() for a in pf_coils_set]
        )

        self.wire = wires

        self.solid = compound

        return compound




from paramak import RotateStraightShape


class VacuumVessel(RotateStraightShape):
    """A cylindrical vessel volume with constant thickness.

    Arguments:
        height (float): height of the vessel.
        inner_radius (float): the inner radius of the vessel.
        thickness (float): thickness of the vessel
    """

    def __init__(
        self,
        height,
        inner_radius,
        thickness,
        **kwargs
    ):
        self.height = height
        self.inner_radius = inner_radius
        self.thickness = thickness
        super().__init__(
            **kwargs
        )

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        if not isinstance(value, (float, int)):
            raise ValueError('VacuumVessel.height must be a number. Not '
                             f'{value}')
        if value <= 0:
            msg = ('VacuumVessel.height must be a positive number above 0. '
                   f'Not {value}')
            raise ValueError(msg)
        self._height = value

    @property
    def inner_radius(self):
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        assert value > 0
        self._inner_radius = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe the
            2D profile of the vessel shape."""
        thickness = self.thickness
        inner_radius = self.inner_radius
        height = self.height

        inner_points = [
            (0, height / 2),
            (inner_radius, height / 2),
            (inner_radius, -height / 2),
            (0, -height / 2),
        ]

        outer_points = [
            (0, height / 2 + thickness),
            (inner_radius + thickness, height / 2 + thickness),
            (inner_radius + thickness, -(height / 2 + thickness)),
            (0, -(height / 2 + thickness)),
        ]
        self.points = inner_points + outer_points[::-1]




import math
import unittest

import paramak
import pytest


class TestHexagonPin(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.HexagonPin(
            length_of_side=5, distance=42., center_point=(0, 0))

    def test_setting_parameters(self):
        """Checks that the default parameters and user parameters are set"""

        assert self.test_shape.length_of_side == 5
        assert self.test_shape.distance == 42.
        assert self.test_shape.center_point == (0, 0)
        assert self.test_shape.name == "hexagon_pin"

    def test_volume(self):
        """Checks the volume against the actual value"""

        length = self.test_shape.length_of_side
        distance = self.test_shape.distance

        hexagon_face_area = (3 * math.sqrt(3) / 2) * math.pow(length, 2)
        # this needs a pytest.approx() as the volumes are not exact
        assert pytest.approx(self.test_shape.volume(),
                             rel=0.1) == hexagon_face_area * distance

    def test_distance_impacts_volume(self):
        """Checks that changing the distance argument results in the
        expected volume change"""

        test_shape_volume = self.test_shape.volume()

        self.test_shape.distance = self.test_shape.distance * 2

        assert pytest.approx(test_shape_volume * 2,
                             rel=0.1) == self.test_shape.volume()

    def test_length_of_sides_impacts_volume(self):
        """Checks that changing the length_of_sides argument results in a the
        expected volume change"""

        test_shape_volume = self.test_shape.volume()

        self.test_shape.length_of_side = self.test_shape.length_of_side * 2

        assert pytest.approx(test_shape_volume * 4,
                             rel=0.1) == self.test_shape.volume()

    def test_areas_are_correct(self):
        """Tests the areas of the faces are the correct sizes"""

        test_shape_areas = self.test_shape.areas

        length = self.test_shape.length_of_side
        distance = self.test_shape.distance

        hexagon_face_area = (3 * math.sqrt(3) / 2) * math.pow(length, 2)

        assert len(test_shape_areas) == 8
        assert test_shape_areas.count(pytest.approx(hexagon_face_area,
                                                    rel=0.1)) == 2
        assert test_shape_areas.count(pytest.approx(length * distance,
                                                    rel=0.1)) == 6




import unittest

import paramak
import pytest


class TestPoloidalFieldCoilFP(unittest.TestCase):

    def test_shape_construction_and_volume(self):
        """Cuts a vessel cylinder with several different size port cutters."""

        test_component = paramak.PoloidalFieldCoilFP(
            corner_points=[(10, 10), (20, 22)])

        assert test_component.volume() == pytest.approx(12063.715789784808)
        assert test_component.corner_points == [(10, 10), (20, 22)]
        assert test_component.solid is not None




import os
import unittest
from pathlib import Path

import paramak


class TestDemo2015Reactor(unittest.TestCase):
    """Tests functionality of the TestDemo2015Reactor class"""

    def test_input_variable_names(self):
        """tests that the number of inputs variables is correct"""

        my_reactor = paramak.EuDemoFrom2015PaperDiagram(number_of_tf_coils=1)
        assert len(my_reactor.input_variables.keys()) == 5
        assert len(my_reactor.input_variable_names) == 5

    def test_plasma_construction(self):
        """Creates the plasma part of the EuDemoFrom2015PaperDiagram and checks
        the contruction runs"""

        my_reactor = paramak.EuDemoFrom2015PaperDiagram(number_of_tf_coils=1)
        plasma = my_reactor.create_plasma()
        assert plasma[0].volume() > 0

    def test_pf_coil_construction(self):
        """Creates the pf coil part of the EuDemoFrom2015PaperDiagram and
        checks the contruction runs"""

        my_reactor = paramak.EuDemoFrom2015PaperDiagram(number_of_tf_coils=1)
        pf_coils = my_reactor.create_pf_coils()
        for coil in pf_coils:
            assert coil.volume() > 0

    def test_vessel_construction(self):
        """Creates the pf coil part of the EuDemoFrom2015PaperDiagram and
        checks the contruction runs"""

        my_reactor = paramak.EuDemoFrom2015PaperDiagram(number_of_tf_coils=1)
        vessel_components = my_reactor.create_vessel_components()
        for component in vessel_components:
            assert component.volume() > 0

    def test_make_demo_2015_reactor(self):
        """Creates a EuDemoFrom2015PaperDiagram reactor and exports the stp
        files checking that each component results in a stp file"""

        output_filenames = [
            'blanket.stp',
            'divertor.stp',
            'outboard_pf_coils.stp',
            'pf_coils_1.stp',
            'pf_coils_2.stp',
            'pf_coils_3.stp',
            'pf_coils_4.stp',
            'pf_coils_5.stp',
            'tf_coil_casing.stp',
            'vacvessel.stp',
            'another.stp'
        ]
        os.system("rm *.stp")
        my_reactor = paramak.EuDemoFrom2015PaperDiagram(number_of_tf_coils=1)
        my_reactor.export_stp(filename=output_filenames)
        for output_filename in output_filenames:
            assert Path(output_filename).exists() is True
        os.system("rm *.stp")

    def test_make_parametric_demo_2015_rector(self):
        """Creates a EuDemoFrom2015PaperDiagram reactor with a non default
        rotation angle and exports the stp files checking that each component
        results in a stp file"""

        output_filenames = [
            'blanket.stp',
            'divertor.stp',
            'outboard_pf_coils.stp',
            'pf_coils_1.stp',
            'pf_coils_2.stp',
            'pf_coils_3.stp',
            'pf_coils_4.stp',
            'pf_coils_5.stp',
            'tf_coil_casing.stp',
            'vacvessel.stp',
            'another.stp'
        ]
        os.system("rm *.stp")
        my_reactor = paramak.EuDemoFrom2015PaperDiagram(
            number_of_tf_coils=1, rotation_angle=90)
        my_reactor.export_stp(filename=output_filenames)
        for output_filename in output_filenames:
            assert Path(output_filename).exists() is True
        os.system("rm *.stp")




import math

from paramak import RotateStraightShape
from paramak.utils import coefficients_of_line_from_points, rotate


class PortCutterRotated(RotateStraightShape):
    """Creates wedges from a central point with angular extent in polar
    direction. To control the width the rotation_angle argument can be used.
    This is useful as a cutting volume for the creation of ports in blankets.

    Args:
        center_point (tuple of floats): the center point where the
            ports are aimed towards, typically the center of the plasma.
        polar_coverage_angle (float): the angular extent of port in the
            polar direction (degrees). Defaults to 10.0.
        polar_placement_angle (float): The angle used when rotating the shape
            on the polar axis. 0 degrees is the outboard equatorial point.
            Defaults to 0.0.
        max_distance_from_center (float): the maximum distance from the center
            point outwards (cm). Default 3000.0.
        fillet_radius (float, optional): If not None, radius (cm) of fillets
            added to all edges. Defaults to 0.0.
        rotation_angle (float, optional): defaults to 10.0.
        name (str, optional): defaults to "port_cutter".
    """

    def __init__(
        self,
        center_point,
        polar_coverage_angle=10.0,
        polar_placement_angle=0.0,
        max_distance_from_center=3000.0,
        fillet_radius=0.0,
        rotation_angle=10.0,
        name="port_cutter",
        **kwargs
    ):

        super().__init__(
            name=name,
            rotation_angle=rotation_angle,
            **kwargs
        )

        self.center_point = center_point
        self.polar_coverage_angle = polar_coverage_angle
        self.polar_placement_angle = polar_placement_angle
        self.max_distance_from_center = max_distance_from_center
        self.fillet_radius = fillet_radius

        self.add_fillet()

    @property
    def polar_coverage_angle(self):
        return self._polar_coverage_angle

    @polar_coverage_angle.setter
    def polar_coverage_angle(self, value):
        if value > 180:
            msg = "polar_coverage_angle must be greater than 180.0"
            raise ValueError(msg)
        self._polar_coverage_angle = value

    @property
    def center_point(self):
        return self._center_point

    @center_point.setter
    def center_point(self, center_point):
        self._center_point = center_point

    @property
    def max_distance_from_center(self):
        return self._max_distance_from_center

    @max_distance_from_center.setter
    def max_distance_from_center(self, value):
        self._max_distance_from_center = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe the
        2D profile of the port cutter."""

        points = [self.center_point]

        outer_point = (self.center_point[0] + self.max_distance_from_center,
                       self.center_point[1])

        outer_point_rotated = rotate(
            self.center_point,
            outer_point,
            math.radians(self.polar_placement_angle))

        outer_point_1 = rotate(
            self.center_point,
            outer_point_rotated,
            math.radians(0.5 * self.polar_coverage_angle))

        outer_point_2 = rotate(
            self.center_point,
            outer_point_rotated,
            math.radians(-0.5 * self.polar_coverage_angle))

        points.append(outer_point_1)
        points.append(outer_point_2)

        self.points = points

    def add_fillet(self):
        """adds fillets to all edges"""
        if self.fillet_radius != 0:
            self.solid = self.solid.edges().fillet(self.fillet_radius)




import math

import numpy as np
from paramak import (RotateMixedShape, distance_between_two_points, extend,
                     rotate)


class ITERtypeDivertor(RotateMixedShape):
    """Creates an ITER-like divertor with inner and outer vertical targets and
    dome

    Args:
        anchors ((float, float), (float, float), optional): xz coordinates of
            points at the top of vertical targets.
            Defaults to ((450, -300), (561, -367)).
        coverages ((float, float), optional): coverages (anticlockwise) in
            degrees of the circular parts of vertical targets.
            Defaults to (90, 180).
        radii ((float, float), optional): radii (cm) of circular parts of the
            vertical targets. Defaults to (50, 25).
        lengths ((float, float), optional): leg length (cm) of the vertical
            targets. Defaults to (78, 87).
        dome (bool, optional): if set to False, the dome will not be created.
            Defaults to True.
        dome_height (float, optional): distance (cm) between the dome base and
            lower points. Defaults to 43.
        dome_length (float, optional): length of the dome. Defaults to 66.
        dome_thickness (float, optional): thickness of the dome.
            Defaults to 10.
        dome_pos (float, optional): relative location of the dome between
            vertical targets (0 inner, 1 outer). Ex: 0.5 will place the dome
            in between the targets. Defaults to 0.5.
        tilts ((float, float), optional): tilt angles (anticlockwise) in
            degrees for the vertical targets. Defaults to (-27, 0).
    """

    def __init__(
        self,
        anchors=((450, -300), (561, -367)),
        coverages=(90, 180),
        radii=(50, 25),
        lengths=(78, 87),
        dome=True,
        dome_height=43,
        dome_length=66,
        dome_thickness=10,
        dome_pos=0.5,
        tilts=(-27, 0),
        **kwargs
    ):

        super().__init__(
            **kwargs
        )

        self.IVT_anchor, self.OVT_anchor = anchors
        self.IVT_coverage, self.OVT_coverage = coverages
        self.IVT_radius, self.OVT_radius = radii
        self.IVT_length, self.OVT_length = lengths
        self.IVT_tilt, self.OVT_tilt = tilts
        self.dome = dome
        self.dome_length = dome_length
        self.dome_height = dome_height
        self.dome_pos = dome_pos
        self.dome_thickness = dome_thickness

    def _create_vertical_target_points(
            self, anchor, coverage, tilt, radius, length):
        """Creates a list of points for a vertical target

        Args:
            anchor (float, float): xz coordinates of point at
                the top of the vertical target.
            coverage (float): coverages (anticlockwise) in degrees of the
                circular part of the vertical target.
            tilt (float): tilt angle (anticlockwise) in
                degrees for the vertical target.
            radius (float): radius (cm) of circular part of the vertical
                target.
            length (float): leg length (cm) of the vertical target.

        Returns:
            list: list of x y coordinates
        """
        points = []
        base_circle_inner = anchor[0] + radius, anchor[1]
        A = rotate(base_circle_inner, anchor, coverage)
        A_prime = rotate(base_circle_inner, anchor, coverage / 2)
        c_coord = (anchor[0], anchor[1] - length)

        A = rotate(anchor, A, tilt)
        A_prime = rotate(anchor, A_prime, tilt)
        c_coord = rotate(anchor, c_coord, tilt)
        # upper inner A
        points.append([A[0], A[1]])
        # A'
        points.append([A_prime[0], A_prime[1]])
        # B
        points.append([anchor[0], anchor[1]])
        # C
        points.append([c_coord[0], c_coord[1]])
        return points

    def _create_dome_points(
        self, c_coord, F, dome_length, dome_height, dome_thickness, dome_pos
    ):
        """Creates a list of points for the dome alongside with their
        connectivity

        Args:
            c_coord (float, float): coordinate of inner end of the dome
            F (float, float): coordinate of outer end of the dome
            dome_length (float): dome length (cm)
            dome_height (float): dome height (cm)
            dome_thickness (float): dome thickness (cm)
            dome_pos (float): position of the dome between the two ends.

        Returns:
            list: list of points with connectivity
                ([[x, z, 'connection_type'], [...]])
        """
        points = []

        dome_base = extend(
            c_coord,
            F,
            dome_pos *
            distance_between_two_points(
                F,
                c_coord))
        dome_lower_point = extend(
            dome_base, rotate(dome_base, c_coord, -math.pi / 2), dome_height
        )

        D_prime = extend(
            dome_base,
            dome_lower_point,
            dome_height +
            dome_thickness)
        D = extend(
            dome_lower_point,
            rotate(dome_lower_point, D_prime, math.pi / 2),
            dome_length / 2,
        )
        E = extend(
            dome_lower_point,
            rotate(dome_lower_point, D_prime, -math.pi / 2),
            dome_length / 2,
        )

        # D
        points.append([D[0], D[1], "circle"])

        # D'
        points.append([D_prime[0], D_prime[1], "circle"])

        # E
        points.append([E[0], E[1], "straight"])
        return points

    def _create_casing_points(self, anchors, c_coord, F, targets_lengths):
        """Creates a list of points for the casing alongside with their
        connectivity

        Args:
            anchors ((float, float), (float, float)): xz coordinates of points
                at the top of vertical targets.
            c_coord (float, float): coordinate of inner end of the dome
            F (float, float): coordinate of outer end of the dome
            targets_lengths (float, float): leg lengths of the vertical targets

        Returns:
            list: list of points with connectivity
                ([[x, z, 'connection_type'], [...]])
        """
        B, G = anchors
        h1, h2 = targets_lengths
        points = []
        # I
        I_ = extend(c_coord, F, distance_between_two_points(F, c_coord) * 1.1)
        points.append([I_[0], I_[1], "straight"])
        # J
        J = extend(G, F, h2 * 1.2)
        points.append([J[0], J[1], "straight"])
        # K
        K = extend(B, c_coord, h1 * 1.2)
        points.append([K[0], K[1], "straight"])
        # L
        L = extend(F, c_coord, distance_between_two_points(F, c_coord) * 1.1)
        points.append([L[0], L[1], "straight"])
        return points

    def find_points(self):
        """Finds the XZ points and connection types (straight and circle) that
        describe the 2D profile of the ITER-like divertor
        """

        # IVT points
        IVT_points = self._create_vertical_target_points(
            self.IVT_anchor,
            math.radians(self.IVT_coverage),
            math.radians(self.IVT_tilt),
            -self.IVT_radius,
            self.IVT_length,
        )
        # add connections
        connections = ["circle"] * 2 + ["straight"] * 2
        for i, connection in enumerate(connections):
            IVT_points[i].append(connection)

        # OVT points
        OVT_points = self._create_vertical_target_points(
            self.OVT_anchor,
            -math.radians(self.OVT_coverage),
            math.radians(self.OVT_tilt),
            self.OVT_radius,
            self.OVT_length,
        )
        # add connections
        connections = ["straight"] + ["circle"] * 2 + ["straight"]
        for i, connection in enumerate(connections):
            OVT_points[i].append(connection)
        # OVT_points need to be fliped for correct connections
        OVT_points = [[float(e[0]), float(e[1]), e[2]]
                      for e in np.flipud(OVT_points)]

        # Dome points
        dome_points = []
        if self.dome:
            dome_points = self._create_dome_points(
                IVT_points[-1][:2],
                OVT_points[0][:2],
                self.dome_length,
                self.dome_height,
                self.dome_thickness,
                self.dome_pos,
            )

        # casing points
        self.casing_points = self._create_casing_points(
            anchors=(self.IVT_anchor, self.OVT_anchor),
            c_coord=IVT_points[-1][:2],
            F=OVT_points[0][:2],
            targets_lengths=(self.IVT_length, self.OVT_length),
        )

        # append all points
        points = IVT_points + dome_points + OVT_points + self.casing_points
        self.points = points




from typing import Optional
from paramak import ExtrudeStraightShape
from paramak.utils import cut_solid


class BlanketCutterParallels(ExtrudeStraightShape):
    """Creates an extruded shape with a parallel rectangular section repeated
    around the reactor. The shape is used to cut other components (eg. blankets
    and firstwalls) in order to create a banana section of the blankets with
    parallel sides.Typically used to divide a blanket into vertical
    sections with a fixed distance between each section.

    Args:
        thickness: extruded distance (cm) of the cutter which translates to
            being the gap size between blankets when the cutter is used to
            segment blankets.
        gap_size: the distance between the inner edges of the two  parallel
            extrusions
        height: height (cm) of the port. Defaults to 2000.0.
        width: width (cm) of the port. Defaults to 2000.0.
        name: Defaults to "blanket_cutter_Parallels".
        azimuth_placement_angle (float or list of floats): Defaults to [0., 36.,
            72., 108., 144., 180., 216., 252., 288., 324.].
    """

    def __init__(
        self,
        thickness: float,
        gap_size: float,
        height: Optional[float] = 2000.,
        width: Optional[float] = 2000.,
        azimuth_placement_angle=[0., 36., 72., 108., 144., 180., 216.,
                                 252., 288., 324.],
        name: Optional[str] = "blanket_cutter_parallels",
        **kwargs
    ) -> None:

        self.main_cutting_shape = \
            ExtrudeStraightShape(
                distance=gap_size / 2.0,
                azimuth_placement_angle=azimuth_placement_angle,
            )
        self.gap_size = gap_size
        self.thickness = thickness
        super().__init__(
            distance=self.distance,
            azimuth_placement_angle=azimuth_placement_angle,
            name=name,
            **kwargs
        )
        self.height = height
        self.width = width

    @property
    def distance(self):
        self.distance = self.gap_size / 2.0 + self.thickness
        return self._distance

    @distance.setter
    def distance(self, value):
        self._distance = value

    @property
    def gap_size(self):
        return self._gap_size

    @gap_size.setter
    def gap_size(self, value):
        self.main_cutting_shape.distance = value / 2.0
        self._gap_size = value

    @property
    def azimuth_placement_angle(self):
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self.main_cutting_shape.azimuth_placement_angle = value
        self._azimuth_placement_angle = value

    def find_points(self):

        points = [
            (0, -self.height / 2),
            (self.width, -self.height / 2),
            (self.width, self.height / 2),
            (0, self.height / 2)
        ]

        self.main_cutting_shape.points = points

        self.points = points[:-1]

    def create_solid(self):
        solid = super().create_solid()
        solid = cut_solid(solid, self.main_cutting_shape)
        self.solid = solid

        return solid




import math
from collections.abc import Iterable
from hashlib import blake2b
from os import fdopen, remove
from pathlib import Path
from shutil import copymode, move
from tempfile import mkstemp
from typing import List, Optional, Tuple, Union

import cadquery as cq
import numpy as np
import plotly.graph_objects as go
from cadquery import importers
from OCP.GCPnts import GCPnts_QuasiUniformDeflection

import paramak


def transform_curve(edge, tolerance: float = 1e-3):
    """Converts a curved edge into a series of straight lines (facetets) with
    the provided tolerance.

    Args:
        edge (cadquery.Wire): The CadQuery wire to redraw as a series of
            straight lines (facet)
        tolerance: faceting toleranceto use when faceting cirles and
            splines. Defaults to 1e-3.

    Returns:
        cadquery.Wire
    """

    curve = edge._geomAdaptor()  # adapt the edge into curve
    start = curve.FirstParameter()
    end = curve.LastParameter()

    points = GCPnts_QuasiUniformDeflection(curve, tolerance, start, end)
    verts = (cq.Vector(points.Value(i + 1)) for i in range(points.NbPoints()))

    return cq.Wire.makePolygon(verts)


def facet_wire(
        wire,
        facet_splines: bool = True,
        facet_circles: bool = True,
        tolerance: float = 1e-3
):
    """Converts specified curved edge types from a wire into a series of
    straight lines (facetets) with the provided tol (tolerance).

    Args:
        wire (cadquery.Wire): The CadQuery wire to select edge from which will
            be redraw as a series of straight lines (facet).
        facet_splines: If True then spline edges will be faceted. Defaults
            to True.
        facet_splines: If True then circle edges will be faceted.Defaults
            to True.
        tolerance: faceting toleranceto use when faceting cirles and
            splines. Defaults to 1e-3.

    Returns:
        cadquery.Wire
    """
    edges = []

    types_to_facet = []
    if facet_splines:
        types_to_facet.append('BSPLINE')
    if facet_circles:
        types_to_facet.append('CIRCLE')

    if isinstance(wire, cq.occ_impl.shapes.Edge):
        # this is for when a edge is passed
        iterable_of_wires = [wire]
    elif isinstance(wire, cq.occ_impl.shapes.Wire):
        # this is for imported stp files
        iterable_of_wires = wire.Edges()
    else:
        # this is for cadquery generated solids
        iterable_of_wires = wire.val().Edges()

    for edge in iterable_of_wires:
        if edge.geomType() in types_to_facet:
            edges.extend(transform_curve(edge, tolerance=tolerance).Edges())
        else:
            edges.append(edge)

    return edges


def coefficients_of_line_from_points(
        point_a: Tuple[float, float], point_b: Tuple[float, float]) -> Tuple[float, float]:
    """Computes the m and c coefficients of the equation (y=mx+c) for
    a straight line from two points.

    Args:
        point_a: point 1 coordinates
        point_b: point 2 coordinates

    Returns:
        m coefficient and c coefficient
    """

    points = [point_a, point_b]
    x_coords, y_coords = zip(*points)
    coord_array = np.vstack([x_coords, np.ones(len(x_coords))]).T
    m, c = np.linalg.lstsq(coord_array, y_coords, rcond=None)[0]
    return m, c


def cut_solid(solid, cutter):
    """
    Performs a boolean cut of a solid with another solid or iterable of solids.

    Args:
        solid Shape: The Shape that you want to cut from
        cutter Shape: The Shape(s) that you want to be the cutting object

    Returns:
        Shape: The original shape cut with the cutter shape(s)
    """

    # Allows for multiple cuts to be applied
    if isinstance(cutter, Iterable):
        for cutting_solid in cutter:
            solid = solid.cut(cutting_solid.solid)
    else:
        solid = solid.cut(cutter.solid)
    return solid


def diff_between_angles(angle_a: float, angle_b: float) -> float:
    """Calculates the difference between two angles angle_a and angle_b

    Args:
        angle_a (float): angle in degree
        angle_b (float): angle in degree

    Returns:
        float: difference between the two angles in degree.
    """

    delta_mod = (angle_b - angle_a) % 360
    if delta_mod > 180:
        delta_mod -= 360
    return delta_mod


def distance_between_two_points(point_a: Tuple[float, float],
                                point_b: Tuple[float, float]) -> float:
    """Computes the distance between two points.

    Args:
        point_a (float, float): X, Y coordinates of the first point
        point_b (float, float): X, Y coordinates of the second point

    Returns:
        float: distance between A and B
    """

    xa, ya = point_a
    xb, yb = point_b
    u_vec = [xb - xa, yb - ya]
    return np.linalg.norm(u_vec)


def extend(point_a: Tuple[float, float], point_b: Tuple[float, float],
           L: float) -> Tuple[float, float]:
    """Creates a point C in (ab) direction so that \\|aC\\| = L

    Args:
        point_a (float, float): X, Y coordinates of the first point
        point_b (float, float): X, Y coordinates of the second point
        L (float): distance AC
    Returns:
        float, float: point C coordinates
    """

    xa, ya = point_a
    xb, yb = point_b
    u_vec = [xb - xa, yb - ya]
    u_vec /= np.linalg.norm(u_vec)

    xc = xa + L * u_vec[0]
    yc = ya + L * u_vec[1]
    return xc, yc


def find_center_point_of_circle(
        point_a: Tuple[float, float],
        point_b: Tuple[float, float],
        point_3: Tuple[float, float]
) -> Tuple[Tuple[float, float], float]:
    """
    Calculates the center and the radius of a circle
    passing through 3 points.
    Args:
        point_a (float, float): point 1 coordinates
        point_b (float, float): point 2 coordinates
        point_3 (float, float): point 3 coordinates
    Returns:
        (float, float), float: center of the circle coordinates or
        None if 3 points on a line are input and the radius
    """

    temp = point_b[0] * point_b[0] + point_b[1] * point_b[1]
    bc = (point_a[0] * point_a[0] + point_a[1] * point_a[1] - temp) / 2
    cd = (temp - point_3[0] * point_3[0] - point_3[1] * point_3[1]) / 2
    det = (point_a[0] - point_b[0]) * (point_b[1] - point_3[1]) - (
        point_b[0] - point_3[0]
    ) * (point_a[1] - point_b[1])

    if abs(det) < 1.0e-6:
        return (None, np.inf)

    # Center of circle
    cx = (bc * (point_b[1] - point_3[1]) -
          cd * (point_a[1] - point_b[1])) / det
    cy = ((point_a[0] - point_b[0]) * cd -
          (point_b[0] - point_3[0]) * bc) / det

    radius = np.sqrt((cx - point_a[0]) ** 2 + (cy - point_a[1]) ** 2)

    return (cx, cy), radius


def intersect_solid(solid, intersecter):
    """
    Performs a boolean intersection of a solid with another solid or iterable of
    solids.
    Args:
        solid Shape: The Shape that you want to intersect
        intersecter Shape: The Shape(s) that you want to be the intersecting object
    Returns:
        Shape: The original shape cut with the intersecter shape(s)
    """

    # Allows for multiple cuts to be applied
    if isinstance(intersecter, Iterable):
        for intersecting_solid in intersecter:
            solid = solid.intersect(intersecting_solid.solid)
    else:
        solid = solid.intersect(intersecter.solid)
    return solid


def rotate(origin: Tuple[float, float], point: Tuple[float, float],
           angle: float):
    """
    Rotate a point counterclockwise by a given angle around a given origin.
    The angle should be given in radians.

    Args:
        origin (float, float): coordinates of origin point
        point (float, float): coordinates of point to be rotated
        angle (float): rotation angle in radians (counterclockwise)
    Returns:
        float, float: rotated point coordinates.
    """

    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


def union_solid(solid, joiner):
    """
    Performs a boolean union of a solid with another solid or iterable of solids

    Args:
        solid (Shape): The Shape that you want to union from
        joiner (Shape): The Shape(s) that you want to form the union with the
            solid
    Returns:
        Shape: The original shape union with the joiner shape(s)
    """

    # Allows for multiple unions to be applied
    if isinstance(joiner, Iterable):
        for joining_solid in joiner:
            solid = solid.union(joining_solid.solid)
    else:
        solid = solid.union(joiner.solid)
    return solid


def calculate_wedge_cut(self):
    """Calculates a wedge cut with the given rotation_angle"""

    if self.rotation_angle == 360:
        return None

    cutting_wedge = paramak.CuttingWedgeFS(self)
    return cutting_wedge


def add_thickness(x: List[float], y: List[float], thickness: float,
                  dy_dx: List[float] = None):
    """Computes outer curve points based on thickness

    Args:
        x (list): list of floats containing x values
        y (list): list of floats containing y values
        thickness (float): thickness of the magnet
        dy_dx (list): list of floats containing the first order
            derivatives

    Returns:
        (list, list): R and Z lists for outer curve points
    """

    if dy_dx is None:
        dy_dx = np.diff(y) / np.diff(x)

    x_outer, y_outer = [], []
    for i in range(len(dy_dx)):
        if dy_dx[i] == float('-inf'):
            nx, ny = -1, 0
        elif dy_dx[i] == float('inf'):
            nx, ny = 1, 0
        else:
            nx = -dy_dx[i]
            ny = 1
        if i != len(dy_dx) - 1:
            if x[i] < x[i + 1]:
                convex = False
            else:
                convex = True

        if convex:
            nx *= -1
            ny *= -1
        # normalize normal vector
        normal_vector_norm = (nx ** 2 + ny ** 2) ** 0.5
        nx /= normal_vector_norm
        ny /= normal_vector_norm
        # calculate outer points
        val_x_outer = x[i] + thickness * nx
        val_y_outer = y[i] + thickness * ny
        x_outer.append(val_x_outer)
        y_outer.append(val_y_outer)

    return x_outer, y_outer


def get_hash(shape, ignored_keys: List = None) -> str:
    """Computes a unique hash value for the shape.

    Args:
        shape (list): The paramak.Shape object to find the hash value for.
        ignored_keys (list, optional): list of shape.__dict__ keys to ignore
            when creating the hash.

    Returns:
        (list, list): R and Z lists for outer curve points
    """

    hash_object = blake2b()
    shape_dict = dict(shape.__dict__)

    if ignored_keys is not None:
        for key in ignored_keys:
            if key in shape_dict.keys():
                shape_dict[key] = None

    hash_object.update(str(list(shape_dict.values())).encode("utf-8"))
    value = hash_object.hexdigest()
    return value


def _replace(filename: str, pattern: str, subst: str) -> None:
    """Opens a file and replaces occurrences of a particular string
        (pattern)with a new string (subst) and overwrites the file.
        Used internally within the paramak to ensure .STP files are
        in units of cm not the default mm.
    Args:
        filename (str): the filename of the file to edit
        pattern (str): the string that should be removed
        subst (str): the string that should be used in the place of the
            pattern string
    """
    # Create temp file
    file_handle, abs_path = mkstemp()
    with fdopen(file_handle, 'w') as new_file:
        with open(filename) as old_file:
            for line in old_file:
                new_file.write(line.replace(pattern, subst))

    # Copy the file permissions from the old file to the new file
    copymode(filename, abs_path)

    # Remove original file
    remove(filename)

    # Move new file
    move(abs_path, filename)


def plotly_trace(
        points: Union[List[Tuple[float, float]], List[Tuple[float, float, float]]],
        mode: str = "markers+lines",
        name: str = None,
        color: Union[Tuple[float, float, float], Tuple[float, float, float, float]] = None
) -> Union[go.Scatter, go.Scatter3d]:
    """Creates a plotly trace representation of the points of the Shape
    object. This method is intended for internal use by Shape.export_html.

    Args:
        points: A list of tuples containing the X, Z points of to add to
            the trace.
        mode: The mode to use for the Plotly.Scatter graph. Options include
            "markers", "lines" and "markers+lines". Defaults to
            "markers+lines"
        name: The name to use in the graph legend color

    Returns:
        plotly trace: trace object
    """

    if color is not None:
        color_list = [i * 255 for i in color]

        if len(color_list) == 3:
            color = "rgb(" + str(color_list).strip("[]") + ")"
        elif len(color_list) == 4:
            color = "rgba(" + str(color_list).strip("[]") + ")"

    if name is None:
        name = "Shape not named"
    else:
        name = name

    text_values = []

    for i, point in enumerate(points):
        text = 'point number= {i} <br> x={point[0]} <br> y= {point[1]}'
        if len(point) == 3:
            text = text + '<br> z= {point[2]} <br>'

        text_values.append(text)

    if all(len(entry) == 3 for entry in points):
        trace = go.Scatter3d(
            x=[row[0] for row in points],
            y=[row[1] for row in points],
            z=[row[2] for row in points],
            mode=mode,
            marker={"size": 3, "color": color},
            name=name
        )

        return trace

    trace = go.Scatter(
        x=[row[0] for row in points],
        y=[row[1] for row in points],
        hoverinfo="text",
        text=text_values,
        mode=mode,
        marker={"size": 5, "color": color},
        name=name,
    )

    return trace


def extract_points_from_edges(
    edges: Union[List[cq.Wire], cq.Wire],
    view_plane: str = 'XZ',
):
    """Extracts points (coordinates) from a CadQuery Edge, optionally projects
    the points to a plane and returns the points.

    Args:
        edges (CadQuery.Wires): The edges to extract points (coordinates from).
        view_plane: The axis to view the points and faceted edges from. The
            options are 'XZ', 'XY', 'YZ', 'YX', 'ZY', 'ZX', 'RZ' and 'XYZ'.
            Defaults to 'RZ'.

    Returns:
        List of Tuples: A list of tuples with float entries for every point
    """

    if isinstance(edges, Iterable):
        list_of_edges = edges
    else:
        list_of_edges = [edges]

    points = []

    for edge in list_of_edges:
        for vertex in edge.Vertices():
            if view_plane == 'XZ':
                points.append((vertex.X, vertex.Z))
            elif view_plane == 'XY':
                points.append((vertex.X, vertex.Y))
            elif view_plane == 'YZ':
                points.append((vertex.Y, vertex.Z))
            elif view_plane == 'YX':
                points.append((vertex.Y, vertex.X))
            elif view_plane == 'ZY':
                points.append((vertex.Z, vertex.Y))
            elif view_plane == 'ZX':
                points.append((vertex.Z, vertex.X))
            elif view_plane == 'RZ':
                xy_coord = math.pow(vertex.X, 2) + math.pow(vertex.Y, 2)
                points.append((math.sqrt(xy_coord), vertex.Z))
            elif view_plane == 'XYZ':
                points.append((vertex.X, vertex.Y, vertex.Z))
            else:
                raise ValueError('view_plane value of ', view_plane,
                                 ' is not supported')
    return points


def load_stp_file(
    filename: str,
    scale_factor: float = 1.
):
    """Loads a stp file and makes the 3D solid and wires available for use.

    Args:
        filename: the filename used to save the html graph.
        scale_factor: a scaling factor to apply to the geometry that can be
            used to increase the size or decrease the size of the geometry.
            Useful when converting the geometry to cm for use in neutronics
            simulations.

    Returns:
        CadQuery.solid, CadQuery.Wires: solid and wires belonging to the object
    """

    part = importers.importStep(str(filename)).val()

    scaled_part = part.scale(scale_factor)
    solid = scaled_part
    wire = scaled_part.Wires()
    return solid, wire


def export_wire_to_html(
    wires,
    filename=None,
    view_plane='RZ',
    facet_splines: bool = True,
    facet_circles: bool = True,
    tolerance: float = 1e-3,
    title=None,
    mode="markers+lines",
):
    """Creates a html graph representation of the points within the wires.
    Edges of certain types (spines and circles) can optionally be faceted.
    If filename provided doesn't end with .html then .html will be added.
    Viewed from the XZ plane

    Args:
        wires (CadQuery.Wire): the wire (edge) or list of wires to plot points
            from and to optionally facet.
        filename: the filename used to save the html graph. If None then no
            html file will saved but a ploty figure will still be returned.
            Defaults to None.
        view_plane: The axis to view the points and faceted edges from. The
            options are 'XZ', 'XY', 'YZ', 'YX', 'ZY', 'ZX', 'RZ' and 'XYZ'.
            Defaults to 'RZ'
        facet_splines: If True then spline edges will be faceted. Defaults to
            True.
        facet_circles: If True then circle edges will be faceted. Defaults to
            True.
        tolerance: faceting toleranceto use when faceting cirles and splines.
            Defaults to 1e-3.
        title: the title of the plotly plot.
        mode: the plotly trace mode to use when plotting the data. Options
            include 'markers+lines', 'markers', 'lines'. Defaults to 'lines'.

    Returns:
        plotly.Figure(): figure object
    """

    fig = go.Figure()
    fig.update_layout(title=title, hovermode="closest")

    if view_plane == 'XYZ':
        fig.update_layout(
            title=title,
            scene_aspectmode='data',
            scene=dict(
                xaxis_title=view_plane[0],
                yaxis_title=view_plane[1],
                zaxis_title=view_plane[2],
            ),
        )
    else:

        fig.update_layout(
            yaxis=dict(scaleanchor="x",
                       scaleratio=1),
            xaxis_title=view_plane[0],
            yaxis_title=view_plane[1]
        )

    if isinstance(wires, list):
        list_of_wires = wires
    else:
        list_of_wires = [wires]

    for counter, wire in enumerate(list_of_wires):

        edges = facet_wire(
            wire=wire,
            facet_splines=facet_splines,
            facet_circles=facet_circles,
            tolerance=tolerance
        )

        points = paramak.utils.extract_points_from_edges(
            edges=edges,
            view_plane=view_plane
        )

        fig.add_trace(
            plotly_trace(
                points=points,
                mode=mode,
                name='edge ' + str(counter)
            )
        )

    for counter, wire in enumerate(list_of_wires):

        if isinstance(wire, cq.occ_impl.shapes.Edge):
            # this is for when an edge is passed
            edges = wire
        elif isinstance(wire, cq.occ_impl.shapes.Wire):
            # this is for imported stp files
            edges = wire.Edges()
        else:
            # this is for cadquery generated solids
            edges = wire.val().Edges()

        points = paramak.utils.extract_points_from_edges(
            edges=edges,
            view_plane=view_plane)

        fig.add_trace(plotly_trace(
            points=points,
            mode="markers",
            name='points on wire ' + str(counter)
        )
        )

    if filename is not None:

        Path(filename).parents[0].mkdir(parents=True, exist_ok=True)

        path_filename = Path(filename)

        if path_filename.suffix != ".html":
            path_filename = path_filename.with_suffix(".html")

        fig.write_html(str(path_filename))

        print("Exported html graph to ", path_filename)

    return fig


def convert_circle_to_spline(
        p_0: Tuple[float, float],
        p_1: Tuple[float, float],
        p_2: Tuple[float, float],
        tolerance: Optional[float] = 0.1
) -> List[Tuple[float, float, str]]:
    """Converts three points on the edge of a circle into a series of points
    on the edge of the circle. This is done by creating a circle edge from the
    the points provided (p_0, p_1, p_2), facets the circle with the provided
    tolerance to extracts the points on the faceted edge and returns them.

    Args:
        p_0: coordinates of the first point
        p_1: coordinates of the second point
        p_2: coordinates of the third point
        tolerance: the precision of the faceting.

    Returns:
        The new points
    """

    # work plane is arbitrarily selected and has no impact of function
    solid = cq.Workplane('XZ').center(0, 0)
    solid = solid.moveTo(p_0[0], p_0[1]).threePointArc(p_1, p_2)
    edge = solid.vals()[0]

    new_edge = paramak.utils.transform_curve(edge, tolerance=tolerance)

    points = paramak.utils.extract_points_from_edges(
        edges=new_edge,
        view_plane='XZ'
    )

    return points


class FaceAreaSelector(cq.Selector):
    """A custom CadQuery selector the selects faces based on their area with a
    tolerance. The following useage example will fillet the faces of an extrude
    shape with an area of 0.5. paramak.ExtrudeStraightShape(points=[(1,1),
    (2,1), (2,2)], distance=5).solid.faces(FaceAreaSelector(0.5)).fillet(0.1)

    Args:
        area (float): The area of the surface to select.
        tolerance (float, optional): The allowable tolerance of the length
            (+/-) while still being selected by the custom selector.
    """

    def __init__(self, area, tolerance=0.1):
        self.area = area
        self.tolerance = tolerance

    def filter(self, object_list):
        """Loops through all the faces in the object checking if the face
        meets the custom selector requirements or not.

        Args:
            object_list (cadquery): The object to filter the faces from.

        Returns:
            object_list (cadquery): The face that match the selector area within
                the specified tolerance.
        """

        new_obj_list = []
        for obj in object_list:
            face_area = obj.Area()

            # Only return faces that meet the requirements
            if face_area > self.area - self.tolerance and face_area < self.area + self.tolerance:
                new_obj_list.append(obj)

        return new_obj_list


class EdgeLengthSelector(cq.Selector):
    """A custom CadQuery selector the selects edges  based on their length with
    a tolerance. The following useage example will fillet the inner edge of a
    rotated triangular shape. paramak.RotateStraightShape(points=[(1,1),(2,1),
    (2,2)]).solid.edges(paramak.EdgeLengthSelector(6.28)).fillet(0.1)

    Args:
        length (float): The length of the edge to select.
        tolerance (float, optional): The allowable tolerance of the length
            (+/-) while still being selected by the custom selector.

    """

    def __init__(self, length: float, tolerance: float = 0.1):
        self.length = length
        self.tolerance = tolerance

    def filter(self, object_list):
        """Loops through all the edges in the object checking if the edge
        meets the custom selector requirements or not.

        Args:
            object_list (cadquery): The object to filter the edges from.

        Returns:
            object_list (cadquery): The edge that match the selector length
                within the specified tolerance.
        """

        new_obj_list = []
        print('filleting edge#')
        for obj in object_list:

            edge_len = obj.Length()

            # Only return edges that meet our requirements
            if edge_len > self.length - self.tolerance and edge_len < self.length + self.tolerance:

                new_obj_list.append(obj)
        print('length(new_obj_list)', len(new_obj_list))
        return new_obj_list




import os
import unittest
from pathlib import Path

import numpy as np
import paramak
import plotly.graph_objects as go
import pytest
from paramak.utils import (EdgeLengthSelector, FaceAreaSelector,
                           extract_points_from_edges, facet_wire,
                           find_center_point_of_circle, plotly_trace)


class TestUtilityFunctions(unittest.TestCase):

    def test_convert_circle_to_spline(self):
        """Tests the conversion of 3 points on a circle into points on a spline
        curve."""

        new_points = paramak.utils.convert_circle_to_spline(
            p_0=(200., 0.),
            p_1=(250., 50.),
            p_2=(200., 100.),
            tolerance=0.2
        )

        # these points can change from 200. to values like 200.00000000000009
        assert pytest.approx(new_points[0][0], abs=0.0000000000001) == 200
        assert pytest.approx(new_points[0][1], abs=0.0000000000001) == 0
        assert pytest.approx(new_points[-1][0], abs=0.0000000000001) == 200
        assert pytest.approx(new_points[-1][1], abs=0.0000000000001) == 100

        new_points_more_details = paramak.utils.convert_circle_to_spline(
            p_0=(200, 0),
            p_1=(250, 50),
            p_2=(200, 100),
            tolerance=0.1
        )

        assert len(new_points_more_details) > len(new_points)

    def test_extract_points_from_edges(self):
        """Extracts points from edges and checks the list returned is the
        correct len and contains the correct types"""

        test_points = [(1, 1), (3, 1), (4, 2)]
        test_shape = paramak.ExtrudeStraightShape(
            points=test_points,
            distance=6,
            workplane='YZ')

        edges = facet_wire(wire=test_shape.wire)

        points = extract_points_from_edges(edges=edges, view_plane='YZ')

        assert len(points) == 6

        for point in points:
            assert len(point) == 2
            assert isinstance(point[0], float)
            assert isinstance(point[1], float)

        points_single_edge = extract_points_from_edges(
            edges=edges[0], view_plane='YZ')

        assert len(points) > len(points_single_edge)
        for point in points_single_edge:
            assert len(point) == 2
            assert isinstance(point[0], float)
            assert isinstance(point[1], float)

        points_single_edge = extract_points_from_edges(
            edges=edges[0], view_plane='XYZ')

        assert len(points) > len(points_single_edge)
        for point in points_single_edge:
            assert len(point) == 3
            assert isinstance(point[0], float)
            assert isinstance(point[1], float)
            assert isinstance(point[2], float)

    def test_trace_creation(self):
        """Creates a plotly trace and checks the type returned"""
        trace = plotly_trace(
            points=[
                (0, 20),
                (20, 0),
                (0, -20)
            ],
            mode='markers+lines',
            color=(10, 10, 10, 0.5)
        )

        assert isinstance(trace, go.Scatter)

    def test_trace_creation_3d(self):
        """Creates a 3d plotly trace and checks the type returned"""
        trace = plotly_trace(
            points=[
                (0, 20, 0),
                (20, 0, 10),
                (0, -20, -10)
            ],
            mode='markers+lines',
            color=(10, 10, 10)
        )

        assert isinstance(trace, go.Scatter3d)

    def test_find_center_point_of_circle(self):
        """passes three points on a circle to the function and checks that the
        radius and center of the circle is calculated correctly"""

        point_a = (0, 20)
        point_b = (20, 0)
        point_3 = (0, -20)

        assert find_center_point_of_circle(
            point_a, point_b, point_3) == (
            (0, 0), 20)

    def test_EdgeLengthSelector_with_fillet_areas(self):
        """tests the filleting of a RotateStraightShape results in an extra
        surface area"""

        test_shape = paramak.RotateStraightShape(
            points=[(1, 1), (2, 1), (2, 2)])

        assert len(test_shape.areas) == 3

        test_shape.solid = test_shape.solid.edges(
            EdgeLengthSelector(6.28)).fillet(0.1)

        assert len(test_shape.areas) == 4

    def test_FaceAreaSelector_with_fillet_areas(self):
        """tests the filleting of a ExtrudeStraightShape"""

        test_shape = paramak.ExtrudeStraightShape(
            distance=5, points=[(1, 1), (2, 1), (2, 2)])

        assert len(test_shape.areas) == 5

        test_shape.solid = test_shape.solid.faces(
            FaceAreaSelector(0.5)).fillet(0.1)

        assert len(test_shape.areas) == 11

    def test_find_center_point_of_circle_zero_det(self):
        """Checks that None is given if det is zero
        """
        point_a = (0, 0)
        point_b = (0, 0)
        point_3 = (0, 0)

        assert find_center_point_of_circle(
            point_a, point_b, point_3) == (
            None, np.inf)




import math
import os
import unittest
from pathlib import Path

import paramak
import pytest


class TestCenterColumnShieldCylinder(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.CenterColumnShieldCylinder(
            height=600, inner_radius=100, outer_radius=200
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a CenterColumnShieldCylinder are correct."""

        assert self.test_shape.rotation_angle == 360

    def test_creation(self):
        """Creates a center column shield using the CenterColumnShieldCylinder
        parametric component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        CenterColumnShieldCylinder component are calculated correctly from the
        parameters given."""

        assert self.test_shape.processed_points == [
            (100, 300, "straight"),
            (200, 300, "straight"),
            (200, -300, "straight"),
            (100, -300, "straight"),
            (100, 300, "straight")
        ]

    def test_points_calculation(self):
        """Checks that the points used to construct the
        CenterColumnShieldCylinder component are calculated correctly from the
        parameters given."""

        assert self.test_shape.points == [
            (100, 300),
            (200, 300),
            (200, -300),
            (100, -300),
        ]

    def test_relative_volume(self):
        """Creates CenterColumnShieldCylinder shapes and checks that their
        relative volumes are correct"""

        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert test_volume == pytest.approx(self.test_shape.volume() * 2)

    def test_absolute_volume(self):
        """Creates a CenterColumnShieldCylinder shape and checks that its
        relative volume is correct"""

        assert self.test_shape.volume() == pytest.approx(
            ((math.pi * (200**2)) - (math.pi * (100**2))) * 600
        )

    def test_absolute_area(self):
        """Creates a CenterColumnShieldCylinder shape and checks that the
        areas of the faces of the solid created are correct"""

        assert len(self.test_shape.areas) == 4
        assert self.test_shape.area == pytest.approx((((math.pi * (200**2)) - (math.pi * (
            100**2))) * 2) + (math.pi * (2 * 200) * 600) + (math.pi * (2 * 100) * 600))
        assert self.test_shape.areas.count(pytest.approx(
            (math.pi * (200**2)) - (math.pi * (100**2)))) == 2
        assert self.test_shape.areas.count(
            pytest.approx(math.pi * (2 * 200) * 600)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(math.pi * (2 * 100) * 600)) == 1

    def test_export_stp_CenterColumnShieldCylinder(self):
        """Creates a CenterColumnShieldCylinder shape and checks that a stp
        file of the shape can be exported using the export_stp method."""

        os.system("rm center_column_shield.stp")
        self.test_shape.export_stp("center_column_shield.stp")
        assert Path("center_column_shield.stp").exists()
        os.system("rm center_column_shield.stp")

    def test_parametric_component_hash_value(self):
        """Creates a parametric component and checks that a cadquery solid with
        a unique hash value is created when .solid is called. Checks that the
        same cadquery solid with the same unique hash value is returned when
        shape.solid is called again after no changes have been made to the
        parametric component. Checks that a new cadquery solid with a new
        unique hash value is constructed when shape.solid is called after
        changes to the parametric component have been made. Checks that the
        hash_value of a parametric component is not updated until a new
        cadquery solid has been created."""

        assert self.test_shape.hash_value is None
        assert self.test_shape.solid is not None
        assert self.test_shape.hash_value is not None
        initial_hash_value = self.test_shape.hash_value
        assert self.test_shape.solid is not None
        assert initial_hash_value == self.test_shape.hash_value
        self.test_shape.height = 120
        assert initial_hash_value == self.test_shape.hash_value
        assert self.test_shape.solid is not None
        assert initial_hash_value != self.test_shape.hash_value

    def test_center_column_shield_cylinder_invalid_parameters_errors(self):
        """Checks that the correct errors are raised when invalid arguments are entered
        as shape parameters."""

        def incorrect_inner_radius():
            self.test_shape.inner_radius = 250

        def incorrect_outer_radius():
            self.test_shape.inner_radius = 100
            self.test_shape.outer_radius = 50

        def incorrect_height():
            self.test_shape.outer_radius = 200
            self.test_shape.height = None

        def incorrect_center_height():
            self.test_shape.center_height = 'string'

        def incorrect_center_height_none():
            self.test_shape.center_height = None

        self.assertRaises(ValueError, incorrect_inner_radius)
        self.assertRaises(ValueError, incorrect_outer_radius)
        self.assertRaises(ValueError, incorrect_height)
        self.assertRaises(TypeError, incorrect_center_height)
        self.assertRaises(TypeError, incorrect_center_height_none)

    def test_center_height_usage(self):
        """Makes a offset cyclinder using center_height that overlaps half of
        the default center_height=0 shape. Cuts shapes and checks volumes.
        """
        test_shape1 = paramak.CenterColumnShieldCylinder(
            inner_radius=3,
            outer_radius=3.3,
            height=5,
            color=(1, 0, 0)
        )

        test_shape2 = paramak.CenterColumnShieldCylinder(
            inner_radius=3,
            outer_radius=3.3,
            height=5,
            center_height=-2.5,
            color=(1, 0, 0)
        )

        test_shape3 = paramak.CenterColumnShieldCylinder(
            inner_radius=3,
            outer_radius=3.3,
            height=5,
            center_height=0,
            color=(1, 0, 0),
            cut=test_shape2
        )

        assert test_shape2.volume() == test_shape1.volume()
        assert pytest.approx(test_shape3.volume()) == 0.5 * \
            test_shape1.volume()




import math
import unittest

import paramak
import pytest


class TestPoloidalFieldCoilCaseSet(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.PoloidalFieldCoilCaseSet(
            heights=[10, 10, 20, 20],
            widths=[10, 10, 20, 40],
            center_points=[(100, 100), (100, 150), (50, 200), (50, 50)],
            casing_thicknesses=[5, 10, 5, 10],
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a PoloidalFieldCoilCaseSet are correct."""

        assert self.test_shape.rotation_angle == 360

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the PoloidalFieldCoilCaseSetFC are
        calculated correctly from the parameters given."""

        assert self.test_shape.processed_points == [
            (105.0, 105.0, 'straight'),
            (105.0, 95.0, 'straight'),
            (95.0, 95.0, 'straight'),
            (95.0, 105.0, 'straight'),
            (105.0, 105.0, 'straight'),
            (110.0, 110.0, 'straight'),
            (110.0, 90.0, 'straight'),
            (90.0, 90.0, 'straight'),
            (90.0, 110.0, 'straight'),
            (110.0, 110.0, 'straight'),
            (105.0, 155.0, 'straight'),
            (105.0, 145.0, 'straight'),
            (95.0, 145.0, 'straight'),
            (95.0, 155.0, 'straight'),
            (105.0, 155.0, 'straight'),
            (115.0, 165.0, 'straight'),
            (115.0, 135.0, 'straight'),
            (85.0, 135.0, 'straight'),
            (85.0, 165.0, 'straight'),
            (115.0, 165.0, 'straight'),
            (60.0, 210.0, 'straight'),
            (60.0, 190.0, 'straight'),
            (40.0, 190.0, 'straight'),
            (40.0, 210.0, 'straight'),
            (60.0, 210.0, 'straight'),
            (65.0, 215.0, 'straight'),
            (65.0, 185.0, 'straight'),
            (35.0, 185.0, 'straight'),
            (35.0, 215.0, 'straight'),
            (65.0, 215.0, 'straight'),
            (70.0, 60.0, 'straight'),
            (70.0, 40.0, 'straight'),
            (30.0, 40.0, 'straight'),
            (30.0, 60.0, 'straight'),
            (70.0, 60.0, 'straight'),
            (80.0, 70.0, 'straight'),
            (80.0, 30.0, 'straight'),
            (20.0, 30.0, 'straight'),
            (20.0, 70.0, 'straight'),
            (80.0, 70.0, 'straight'),
            (105.0, 105.0, 'straight')
        ]

    def test_points_calculation(self):
        """Checks that the points used to construct the PoloidalFieldCoilCaseSetFC are
        calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (105.0, 105.0),
            (105.0, 95.0),
            (95.0, 95.0),
            (95.0, 105.0),
            (105.0, 105.0),
            (110.0, 110.0),
            (110.0, 90.0),
            (90.0, 90.0),
            (90.0, 110.0),
            (110.0, 110.0),
            (105.0, 155.0),
            (105.0, 145.0),
            (95.0, 145.0),
            (95.0, 155.0),
            (105.0, 155.0),
            (115.0, 165.0),
            (115.0, 135.0),
            (85.0, 135.0),
            (85.0, 165.0),
            (115.0, 165.0),
            (60.0, 210.0),
            (60.0, 190.0),
            (40.0, 190.0),
            (40.0, 210.0),
            (60.0, 210.0),
            (65.0, 215.0),
            (65.0, 185.0),
            (35.0, 185.0),
            (35.0, 215.0),
            (65.0, 215.0),
            (70.0, 60.0),
            (70.0, 40.0),
            (30.0, 40.0),
            (30.0, 60.0),
            (70.0, 60.0),
            (80.0, 70.0),
            (80.0, 30.0),
            (20.0, 30.0),
            (20.0, 70.0),
            (80.0, 70.0),
        ]

    def test_creation(self):
        """Creates a set of pf coils using the PoloidalFieldCoilCaseSet
        parametric component and passing all required args, and checks
        that a solid with the correct number of solids is created."""

        assert self.test_shape.solid is not None
        assert len(self.test_shape.solid.Solids()) == 4

    def test_creation_with_zero_thickness(self):
        """Creates a set of pf coils using the PoloidalFieldCoilCaseSet
        parametric component and passing a 0 entry into the casing_thicknesses
        list, and checks that a solid with the correct number of solids is
        created."""

        self.test_shape.casing_thicknesses = [5, 0, 10, 10]

        assert self.test_shape.solid is not None
        assert len(self.test_shape.solid.Solids()) == 3

    def test_absolute_volume(self):
        """Creates a set of pf coils using the PoloidalFieldCoilCaseSet
        parametric component and checks that the volume is correct."""

        assert self.test_shape.volume() == pytest.approx((((20 * 5 * 2) +
                                                           (10 * 5 * 2)) * math.pi * 2 * 100) + (((30 * 10 * 2) +
                                                                                                  (10 * 10 * 2)) * math.pi * 2 * 100) + (((30 * 5 * 2) +
                                                                                                                                          (20 * 5 * 2)) * math.pi * 2 * 50) + (((60 * 10 * 2) +
                                                                                                                                                                                (20 * 10 * 2)) * math.pi * 2 * 50))

    def test_absolute_areas(self):
        """Creates a set of pf coils using the PoloidalFieldCoilCaseSet
        parametric component and checks that the areas are correct"""

        assert len(self.test_shape.areas) == 32
        assert len(set([round(i) for i in self.test_shape.areas])) == 16
        assert self.test_shape.areas.count(
            pytest.approx(10 * math.pi * 2 * 100)) == 6
        assert self.test_shape.areas.count(
            pytest.approx(40 * math.pi * 2 * 50)) == 4
        assert self.test_shape.areas.count(
            pytest.approx(30 * math.pi * 2 * 100)) == 4
        assert self.test_shape.areas.count(
            pytest.approx(30 * math.pi * 2 * 50)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(10 * math.pi * 2 * 105)) == 3
        assert self.test_shape.areas.count(
            pytest.approx(10 * math.pi * 2 * 95)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 110)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 90)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(30 * math.pi * 2 * 115)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(30 * math.pi * 2 * 85)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 60)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 40)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(30 * math.pi * 2 * 65)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 70)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(20 * math.pi * 2 * 30)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(40 * math.pi * 2 * 80)) == 1

    def test_PoloidalFieldCoilCaseSet_incorrect_thicknesses_1(self):
        """Checks that an error is raised when a PoloidalFieldCoilCaseSet is made
        with the wrong number of casing thicknesses."""

        def make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_1():
            self.test_shape.casing_thicknesses = [5, 5, 10]
            self.test_shape.solid

        self.assertRaises(
            ValueError,
            make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_1
        )

    def test_PoloidalFieldCoil_incorrect_thicknesses_2(self):
        """Checks that an error is raised when a PoloidalFieldCoilCaseSet is made
        with invalid casing thicknesses."""

        def make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_2():
            self.test_shape.casing_thicknesses = [5, 5, 5, 'ten']

        self.assertRaises(
            ValueError,
            make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_2
        )

    def test_PoloidalFieldCoil_incorrect_thicknesses_3(self):
        """Checks that an error is raised when a PoloidalFieldCoilCaseSet is made
        with invalid casing thicknesses."""

        def make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_3():
            self.test_shape.casing_thicknesses = "ten"

        self.assertRaises(
            ValueError,
            make_PoloidalFieldCoilCaseSet_incorrect_thicknesses_3
        )




import math

from paramak import RotateStraightShape
from paramak.utils import rotate


class RotatedTrapezoid(RotateStraightShape):
    """Creates a rotated trapezoid (truncated triangle) shape.

    Args:
        length_1 (float): the length of the top parrallel edge of the trapezoid
            (cm).
        length_2 (float): the length of the base parrallel edge of the
            trapezoid (cm).
        length_3 (float): the height of the trapezoid, the distances from top
            to base (cm).
        pivot_point ((float, float)): the coordinates of the center of
            rotation (x,z). The piviot point is located in the center of the
            length_1 edge (cm).
        pivot_angle (float, optional): the angle (in degrees) to pivot (rotate)
            the shape by around the pivot point. Defaults to 0.
        name (str, optional): defaults to "rotated_trapezoid".
    """

    def __init__(
        self,
        length_1,
        length_2,
        length_3,
        pivot_point,
        pivot_angle=0.,
        name="rotated_trapezoid",
        **kwargs
    ):

        super().__init__(
            name=name,
            **kwargs
        )

        self.length_1 = length_1
        self.length_2 = length_2
        self.length_3 = length_3
        self.pivot_point = pivot_point
        self.pivot_angle = pivot_angle

    @property
    def pivot_point(self):
        return self._pivot_point

    @pivot_point.setter
    def pivot_point(self, pivot_point):
        self._pivot_point = pivot_point

    @property
    def length_1(self):
        return self._length_1

    @length_1.setter
    def length_1(self, length_1):
        self._length_1 = length_1

    @property
    def length_2(self):
        return self._length_2

    @length_2.setter
    def length_2(self, length_2):
        self._length_2 = length_2

    @property
    def length_3(self):
        return self._length_3

    @length_3.setter
    def length_3(self, length_3):
        self._length_3 = length_3

    @property
    def pivot_angle(self):
        return self._pivot_angle

    @pivot_angle.setter
    def pivot_angle(self, pivot_angle):
        self._pivot_angle = pivot_angle

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the trapezoid shape."""

        non_rotated_points = [
            (
                self.pivot_point[0] + self.length_1 / 2.0,
                self.pivot_point[1]
            ),
            (
                self.pivot_point[0] - self.length_1 / 2.0,
                self.pivot_point[1]
            ),
            (
                self.pivot_point[0] - self.length_2 / 2.0,
                self.pivot_point[1] - self.length_3
            ),
            (
                self.pivot_point[0] + self.length_2 / 2.0,
                self.pivot_point[1] - self.length_3
            ),
        ]

        points = []

        for point in non_rotated_points:
            x, y = rotate(
                origin=self.pivot_point,
                point=point,
                angle=math.radians(self.pivot_angle)
            )
            points.append((x, y))

        self.points = points




import unittest

import paramak


class TestCenterColumnShieldPlasmaHyperbola(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.CenterColumnShieldPlasmaHyperbola(
            height=800, inner_radius=100, mid_offset=40, edge_offset=30
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a CenterColumnShieldPlasmaHyperbola are
        correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.major_radius == 450
        assert self.test_shape.minor_radius == 150
        assert self.test_shape.triangularity == 0.55
        assert self.test_shape.elongation == 2

    def test_creation(self):
        """Creates a center column shield using the
        CenterColumnShieldPlasmaHyperbola parametric component and checks that
        a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_points_calculation(self):
        """Checks that the points used to construct the
        CenterColumnShieldPlasmaHyperbola component are calculated correctly
        from the parameters given."""

        assert self.test_shape.points == [
            (100, 0, 'straight'),
            (100, 400.0, 'straight'),
            (337.5, 400.0, 'straight'),
            (337.5, 300.0, 'spline'),
            (260.0, 0.0, 'spline'),
            (337.5, -300.0, 'straight'),
            (337.5, -400.0, 'straight'),
            (100, -400.0, 'straight'),
        ]

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        CenterColumnShieldPlasmaHyperbola component are calculated correctly
        from the parameters given."""

        assert self.test_shape.processed_points == [
            (100, 0, 'straight'),
            (100, 400.0, 'straight'),
            (337.5, 400.0, 'straight'),
            (337.5, 300.0, 'spline'),
            (260.0, 0.0, 'spline'),
            (337.5, -300.0, 'straight'),
            (337.5, -400.0, 'straight'),
            (100, -400.0, 'straight'),
            (100, 0, 'straight')
        ]

    def test_invalid_parameters_errors(self):
        """Checks that the correct errors are raised when invalid arguments are input as
        shape parameters."""

        def incorrect_inner_radius():
            self.test_shape.inner_radius = 601
            self.test_shape.solid

        def incorrect_height():
            self.test_shape.inner_radius = 100
            self.test_shape.height = 300
            self.test_shape.solid

        self.assertRaises(ValueError, incorrect_inner_radius)
        self.assertRaises(ValueError, incorrect_height)

    def test_faces(self):
        """Creates a center column shield using the CenterColumnShieldPlasmaHyperbola
        parametric component and checks that a solid with the correct number of
        faces is created"""

        assert len(self.test_shape.areas) == 6
        assert len(set([round(i) for i in self.test_shape.areas])) == 4

        self.test_shape.rotation_angle = 180
        assert len(self.test_shape.areas) == 8
        assert len(set([round(i) for i in self.test_shape.areas])) == 5




import collections
import json
from collections import Counter
from collections.abc import Iterable
from pathlib import Path
from typing import List, Optional, Tuple, Union

import cadquery as cq
import matplotlib.pyplot as plt
from cadquery import Compound, exporters

import paramak
from paramak.utils import _replace, get_hash


class Reactor:
    """The Reactor object allows shapes and components to be added and then
    collective operations to be performed on them. Combining all the shapes is
    required for creating images of the whole reactor and creating a Graveyard
    (bounding box) that is useful for neutronics simulations.

    Args:
        shapes_and_components: list of paramak.Shape objects
        graveyard_size: The dimension of cube shaped the graveyard region used
            by DAGMC. This attribute is used preferentially over
            graveyard_offset.
        graveyard_offset: The distance between the graveyard and the largest
            shape. If graveyard_size is set the this is ignored.
        largest_shapes: Identifying the shape(s) with the largest size in each
            dimension (x,y,z) can speed up the production of the graveyard.
            Defaults to None which finds the largest shapes by looping through
            all the shapes and creating bounding boxes. This can be slow and
            that is why the user is able to provide a subsection of shapes to
            use when calculating the graveyard dimensions.
    """

    def __init__(
            self,
            shapes_and_components: List[paramak.Shape] = [],
            graveyard_size: Optional[float] = 20_000,
            graveyard_offset: Optional[float] = None,
            largest_shapes: Optional[List[paramak.Shape]] = None,
    ):

        self.shapes_and_components = shapes_and_components
        self.graveyard_offset = graveyard_offset
        self.graveyard_size = graveyard_size
        self.largest_shapes = largest_shapes

        self.input_variable_names = [
            # 'shapes_and_components', commented out to avoid calculating solids
            'graveyard_size',
            'graveyard_offset',
            'largest_shapes',
        ]

        self.stp_filenames = []
        self.stl_filenames = []

        self.tet_meshes = []
        self.graveyard = None
        self.solid = None

        self.reactor_hash_value = None

    @property
    def input_variables(self):
        all_input_variables = {}
        for name in self.input_variable_names:
            all_input_variables[name] = getattr(self, name)
        return all_input_variables

    @property
    def graveyard_size(self):
        return self._graveyard_size

    @graveyard_size.setter
    def graveyard_size(self, value):
        if value is None:
            self._graveyard_size = None
        elif not isinstance(value, (float, int)):
            raise TypeError("graveyard_size must be a number")
        elif value < 0:
            raise ValueError("graveyard_size must be positive")
        self._graveyard_size = value

    @property
    def graveyard_offset(self):
        return self._graveyard_offset

    @graveyard_offset.setter
    def graveyard_offset(self, value):
        if value is None:
            self._graveyard_offset = None
        elif not isinstance(value, (float, int)):
            raise TypeError("graveyard_offset must be a number")
        elif value < 0:
            raise ValueError("graveyard_offset must be positive")
        self._graveyard_offset = value

    @property
    def largest_dimension(self):
        """Calculates a bounding box for the Reactor and returns the largest
        absolute value of the largest dimension of the bounding box"""
        largest_dimension = 0

        if self.largest_shapes is None:
            shapes_to_bound = self.shapes_and_components
        else:
            shapes_to_bound = self.largest_shapes

        for component in shapes_to_bound:
            largest_dimension = max(
                largest_dimension,
                component.largest_dimension)
        # self._largest_dimension = largest_dimension
        return largest_dimension

    @largest_dimension.setter
    def largest_dimension(self, value):
        self._largest_dimension = value

    @property
    def tet_meshes(self):
        values = []
        for shape_or_component in self.shapes_and_components:
            values.append(shape_or_component.tet_mesh)
        return values

    @tet_meshes.setter
    def tet_meshes(self, value):
        self._tet_meshes = value

    @property
    def largest_shapes(self):
        return self._largest_shapes

    @largest_shapes.setter
    def largest_shapes(self, value):
        if not isinstance(value, (list, tuple, type(None))):
            raise ValueError('paramak.Reactor.largest_shapes should be a '
                             'list of paramak.Shapes')
        self._largest_shapes = value

    @property
    def shapes_and_components(self):
        """Adds a list of parametric shape(s) and or parametric component(s)
        to the Reactor object. This allows collective operations to be
        performed on all the shapes in the reactor."""
        if hasattr(self, "create_solids"):
            ignored_keys = ["reactor_hash_value"]
            if get_hash(self, ignored_keys) != self.reactor_hash_value:
                self.create_solids()
                self.reactor_hash_value = get_hash(self, ignored_keys)
        return self._shapes_and_components

    @shapes_and_components.setter
    def shapes_and_components(self, value):
        if not isinstance(value, (Iterable, str)):
            raise ValueError("shapes_and_components must be a list")
        self._shapes_and_components = value

    @property
    def graveyard_offset(self):
        return self._graveyard_offset

    @graveyard_offset.setter
    def graveyard_offset(self, value):
        if value is None:
            self._graveyard_offset = None
        elif not isinstance(value, (float, int)):
            raise TypeError("graveyard_offset must be a number")
        elif value < 0:
            raise ValueError("graveyard_offset must be positive")
        self._graveyard_offset = value

    @property
    def solid(self):
        """This combines all the parametric shapes and components in the
        reactor object.
        """

        list_of_cq_vals = []
        for shape_or_compound in self.shapes_and_components:
            if isinstance(
                shape_or_compound.solid,
                (cq.occ_impl.shapes.Shape,
                    cq.occ_impl.shapes.Compound)):
                for solid in shape_or_compound.solid.Solids():
                    list_of_cq_vals.append(solid)
            else:
                list_of_cq_vals.append(shape_or_compound.solid.val())

        compound = cq.Compound.makeCompound(list_of_cq_vals)

        return compound

    @property
    def name(self):
        """Returns a list of names of the individual Shapes that make up the
        reactor """

        all_names = []
        for shape in self.shapes_and_components:
            all_names.append(shape.name)

        return all_names

    @ solid.setter
    def solid(self, value):
        self._solid = value

    def show(self, default_edgecolor: Tuple[float, float, float] = (0, 0, 0)):
        """Shows / renders the CadQuery the 3d object in Jupyter Lab. Imports
        show from jupyter_cadquery.cadquery and returns show(Reactor.solid)

        Args:
            default_edgecolor: the color to use for the edges, passed to
                jupyter_cadquery.cadquery show. Tuple of three values expected
                individual values in the tuple should be floats between 0. and
                1.

        Returns:
            jupyter_cadquery.cadquery.show object
        """

        try:
            from jupyter_cadquery.cadquery import Part, PartGroup, show
        except ImportError:
            msg = (
                'To use Reactor.show() you must install jupyter_cadquery. To'
                'install jupyter_cadquery type "pip install jupyter_cadquery"'
                ' in the terminal')
            raise ImportError(msg)

        parts = []
        for shape_or_compound in self.shapes_and_components:

            if shape_or_compound.name is None:
                name = 'Shape.name not set'
            else:
                name = shape_or_compound.name

            scaled_color = [int(i * 255) for i in shape_or_compound.color[0:3]]
            scaled_edge_color = [int(i * 255) for i in default_edgecolor[0:3]]
            if isinstance(
                    shape_or_compound.solid,
                    (cq.occ_impl.shapes.Shape, cq.occ_impl.shapes.Compound)):
                for i, solid in enumerate(shape_or_compound.solid.Solids()):
                    parts.append(
                        Part(
                            solid,
                            name=f"{name}{i}",
                            color=scaled_color))
            else:
                parts.append(
                    Part(
                        shape_or_compound.solid.val(),
                        name=f"{name}",
                        color=scaled_color))

        return show(PartGroup(parts), default_edgecolor=scaled_edge_color)

    def export_stp(
            self,
            filename: Union[List[str], str],
            mode: Optional[str] = 'solid',
            units: Optional[str] = 'mm',
    ) -> Union[List[str], str]:
        """Exports the 3D reactor model as a stp file or files.

        Args:
            filename: Accepts a single filename as a string which exports the
                full reactor model to a single file. Alternativley filename can
                also accept a list of strings where each string is the filename
                of the the individual shapes that make it up. This will result
                in separate files for each shape in the reactor.
            mode: the object to export can be either 'solid' which exports 3D
                solid shapes or the 'wire' which exports the wire edges of the
                shape.
            units: the units of the stp file, options are 'cm' or 'mm'.
                Default is mm.
        Returns:
            The stp filename(s) created
        """

        if isinstance(filename, str):

            # exports a single file for the whole model
            assembly = cq.Assembly(name='reactor')
            for entry in self.shapes_and_components:
                if entry.color is None:
                    assembly.add(entry.solid)
                else:
                    assembly.add(entry.solid, color=cq.Color(*entry.color))

            assembly.save(filename, exportType='STEP')

            if units == 'cm':
                _replace(
                    filename,
                    'SI_UNIT(.MILLI.,.METRE.)',
                    'SI_UNIT(.CENTI.,.METRE.)')

            return [filename]

       # exports the reactor solid as a separate stp files
        if len(filename) != len(self.shapes_and_components):
            msg = (f'The Reactor contains {len(self.shapes_and_components)} '
                   f'Shapes and {len(filename)} filenames have be provided')
            raise ValueError(msg)

        for stp_filename, entry in zip(filename, self.shapes_and_components):

            entry.export_stp(
                filename=stp_filename,
                mode=mode,
                units=units,
                verbose=False,
            )

            if units == 'cm':
                _replace(
                    stp_filename,
                    'SI_UNIT(.MILLI.,.METRE.)',
                    'SI_UNIT(.CENTI.,.METRE.)')

        return filename

    def export_brep(
        self,
        filename
    ):
        """Exports a brep file for the Reactor.solid.

        Args:
            filename: the filename of exported the brep file.
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".brep":
            msg = "When exporting a brep file the filename must end with .brep"
            raise ValueError(msg)

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        self.solid.exportBrep(str(path_filename))
        # alternative method is to use BRepTools that might support imprinting
        # and merging https://github.com/CadQuery/cadquery/issues/449

        return str(path_filename)

    def export_stl(
            self,
            filename: Union[List[str], str],
            tolerance: Optional[float] = 0.001,
            angular_tolerance: Optional[float] = 0.1,
    ) -> List[str]:
        """Writes stl files (CAD geometry) for each Shape object in the reactor

        Args:
            tolerance (float):  the precision of the faceting
            include_graveyard: specify if the graveyard will be included or
                not. If True the the Reactor.make_graveyard will be called
                using Reactor.graveyard_size and Reactor.graveyard_offset
                attribute values.

        Returns:
            list: a list of stl filenames created
        """

        if isinstance(filename, str):

            path_filename = Path(filename)

            if path_filename.suffix != ".stl":
                path_filename = path_filename.with_suffix(".stl")

            path_filename.parents[0].mkdir(parents=True, exist_ok=True)

            # add an include_graveyard that add graveyard if requested
            exporters.export(self.solid, str(path_filename), exportType='STL',
                             tolerance=tolerance,
                             angularTolerance=angular_tolerance)
            return str(path_filename)

        # exports the reactor solid as a separate stl files
        if len(filename) != len(self.shapes_and_components):
            msg = (f'The Reactor contains {len(self.shapes_and_components)} '
                   f'Shapes and {len(filename)} filenames have be provided')
            raise ValueError(msg)

        for stl_filename, entry in zip(filename, self.shapes_and_components):

            entry.export_stl(
                filename=stl_filename,
                tolerance=tolerance,
                verbose=False,
            )

        return filename

    def make_sector_wedge(
            self,
            height: Optional[float] = None,
            radius: Optional[float] = None,
            rotation_angle: Optional[float] = None,
    ) -> Union[paramak.Shape, None]:
        """Creates a rotated wedge shaped object that is useful for creating
        sector models in DAGMC where reflecting surfaces are needed. If the
        rotation

        Args:
            height: The height of the rotated wedge. If None then the
                largest_dimension of the model will be used.
            radius: The radius of the rotated wedge. If None then the
                largest_dimension of the model will be used
            rotation_angle: The rotation angle of the wedge will be the
                inverse of the sector

        Returns:
            the paramak.Shape object created
        """

        if rotation_angle is None:
            if hasattr(self, 'rotation_angle'):
                rotation_angle = self.rotation_angle
            if rotation_angle is None:
                Warning('No sector_wedge can be made as rotation_angle'
                        ' or Reactor.rotation_angle have not been set')
                return None

        if rotation_angle > 360:
            Warning(
                'No wedge can be made for a rotation angle of 360 or above')
            return None

        if rotation_angle == 360:
            print('No sector wedge made as rotation angle is 360')
            return None

        if height is None:
            height = self.largest_dimension * 2

        if radius is None:
            radius = self.largest_dimension * 2

        sector_cutting_wedge = paramak.CuttingWedge(
            height=height,
            radius=radius,
            rotation_angle=360 - rotation_angle,
            surface_reflectivity=True,
            azimuth_placement_angle=rotation_angle,
        )

        self.sector_wedge = sector_cutting_wedge

        return sector_cutting_wedge

    def export_svg(
            self,
            filename: Optional[str] = 'reactor.svg',
            projectionDir: Tuple[float, float, float] = (-1.75, 1.1, 5),
            width: Optional[float] = 1000,
            height: Optional[float] = 800,
            marginLeft: Optional[float] = 120,
            marginTop: Optional[float] = 100,
            strokeWidth: Optional[float] = None,
            strokeColor: Optional[Tuple[int, int, int]] = (0, 0, 0),
            hiddenColor: Optional[Tuple[int, int, int]] = (100, 100, 100),
            showHidden: Optional[bool] = False,
            showAxes: Optional[bool] = False,
    ) -> str:
        """Exports an svg file for the Reactor.solid. If the filename provided
        doesn't end with .svg it will be added.

        Args:
            filename: the filename of the svg file to be exported. Defaults to
                "reactor.svg".
            projectionDir: The direction vector to view the geometry from
                (x, y, z). Defaults to (-1.75, 1.1, 5)
            width: the width of the svg image produced in pixels. Defaults to
                1000
            height: the height of the svg image produced in pixels. Defaults to
                800
            marginLeft: the number of pixels between the left edge of the image
                and the start of the geometry.
            marginTop: the number of pixels between the top edge of the image
                and the start of the geometry.
            strokeWidth: the width of the lines used to draw the geometry.
                Defaults to None which automatically selects an suitable width.
            strokeColor: the color of the lines used to draw the geometry in
                RGB format with each value between 0 and 255. Defaults to
                (0, 0, 0) which is black.
            hiddenColor: the color of the lines used to draw the geometry in
                RGB format with each value between 0 and 255. Defaults to
                (100, 100, 100) which is light grey.
            showHidden: If the edges obscured by geometry should be included in
                the diagram. Defaults to False.
            showAxes: If the x, y, z axis should be included in the image.
                Defaults to False.

        Returns:
            str: the svg filename created
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".svg":
            path_filename = path_filename.with_suffix(".svg")

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        opt = {
            "width": width,
            "height": height,
            "marginLeft": marginLeft,
            "marginTop": marginTop,
            "showAxes": showAxes,
            "projectionDir": projectionDir,
            "strokeColor": strokeColor,
            "hiddenColor": hiddenColor,
            "showHidden": showHidden
        }

        if strokeWidth is not None:
            opt["strokeWidth"] = strokeWidth

        exporters.export(self.solid, str(path_filename), exportType='SVG',
                         opt=opt)

        print("Saved file as ", path_filename)

        return str(path_filename)

    def export_stp_graveyard(
            self,
            filename: Optional[str] = "graveyard.stp",
            graveyard_size: Optional[float] = None,
            graveyard_offset: Optional[float] = None,
    ) -> str:
        """Writes a stp file (CAD geometry) for the reactor graveyard. This
        is needed for DAGMC simulations. This method also calls
        Reactor.make_graveyard() with the graveyard_size and graveyard_size
        values.

        Args:
            filename (str): the filename for saving the stp file. Appends
                .stp to the filename if it is missing.
            graveyard_size: directly sets the size of the graveyard. Defaults
                to None which then uses the Reactor.graveyard_size attribute.
            graveyard_offset: the offset between the largest edge of the
                geometry and inner bounding shell created. Defaults to None
                which then uses Reactor.graveyard_offset attribute.

        Returns:
            str: the stp filename created
        """

        graveyard = self.make_graveyard(
            graveyard_offset=graveyard_offset,
            graveyard_size=graveyard_size,
        )

        path_filename = Path(filename)

        if path_filename.suffix != ".stp":
            path_filename = path_filename.with_suffix(".stp")

        graveyard.export_stp(filename=str(path_filename))

        return str(path_filename)

    def make_graveyard(
            self,
            graveyard_size: Optional[float] = None,
            graveyard_offset: Optional[float] = None,
    ) -> paramak.Shape:
        """Creates a graveyard volume (bounding box) that encapsulates all
        volumes. This is required by DAGMC when performing neutronics
        simulations. The graveyard size can be ascertained in two ways. Either
        the size can be set directly using the graveyard_size which is the
        quickest method. Alternativley the graveyard can be automatically sized
        to the geometry by setting a graveyard_offset value. If both options
        are set then the method will default to using the graveyard_size
        preferentially.

        Args:
            graveyard_size: directly sets the size of the graveyard. Defaults
                to None which then uses the Reactor.graveyard_size attribute.
            graveyard_offset: the offset between the largest edge of the
                geometry and inner bounding shell created. Defaults to None
                which then uses Reactor.graveyard_offset attribute.

        Returns:
            CadQuery solid: a shell volume that bounds the geometry, referred
            to as a graveyard in DAGMC
        """

        if graveyard_size is not None:
            graveyard_size_to_use = graveyard_size

        elif self.graveyard_size is not None:
            graveyard_size_to_use = self.graveyard_size

        elif graveyard_offset is not None:
            self.solid
            graveyard_size_to_use = self.largest_dimension * 2 + graveyard_offset * 2

        elif self.graveyard_offset is not None:
            self.solid
            graveyard_size_to_use = self.largest_dimension * 2 + self.graveyard_offset * 2

        else:
            raise ValueError("the graveyard_size, Reactor.graveyard_size, \
                graveyard_offset and Reactor.graveyard_offset are all None. \
                Please specify at least one of these attributes or arguments")

        graveyard_shape = paramak.HollowCube(
            length=graveyard_size_to_use,
            name="graveyard",
        )

        self.graveyard = graveyard_shape

        return graveyard_shape

    def export_2d_image(
            self,
            filename: Optional[str] = "2d_slice.png",
            xmin: Optional[float] = 0.0,
            xmax: Optional[float] = 900.0,
            ymin: Optional[float] = -600.0,
            ymax: Optional[float] = 600.0,
    ) -> str:
        """Creates a 2D slice image (png) of the reactor.

        Args:
            filename (str): output filename of the image created

        Returns:
            str: png filename created
        """

        path_filename = Path(filename)

        if path_filename.suffix != ".png":
            path_filename = path_filename.with_suffix(".png")

        path_filename.parents[0].mkdir(parents=True, exist_ok=True)

        fig, ax = plt.subplots()

        # creates indvidual patches for each Shape which are combined together
        for entry in self.shapes_and_components:
            patch = entry._create_patch()
            ax.add_collection(patch)

        ax.axis("equal")
        ax.set(xlim=(xmin, xmax), ylim=(ymin, ymax))
        ax.set_aspect("equal", "box")

        Path(filename).parent.mkdir(parents=True, exist_ok=True)
        plt.savefig(filename, dpi=100)
        plt.close()

        print("\n saved 2d image to ", str(path_filename))

        return str(path_filename)

    def export_html_3d(
            self,
            filename: Optional[str] = "reactor_3d.html",
    ) -> str:
        """Saves an interactive 3d html view of the Reactor to a html file.

        Args:
            filename: the filename used to save the html graph. Defaults to
                reactor_3d.html

        Returns:
            str: filename of the created html file
        """

        from ipywidgets.embed import embed_minimal_html

        view = self.show()

        if view is None:
            return None

        embed_minimal_html(
            filename,
            views=[view.cq_view.renderer],
            title='Renderer'
        )

        return filename

    def export_html(
            self,
            filename: Optional[str] = "reactor.html",
            facet_splines: Optional[bool] = True,
            facet_circles: Optional[bool] = True,
            tolerance: Optional[float] = 1.,
            view_plane: Optional[str] = 'RZ'):
        """Creates a html graph representation of the points for the Shape
        objects that make up the reactor. Shapes are colored by their .color
        property. Shapes are also labelled by their .name. If filename provided
        doesn't end with .html then .html will be added.

        Args:
            filename: the filename used to save the html graph. Defaults to
                reactor.html
            facet_splines: If True then spline edges will be faceted. Defaults
                to True.
            facet_circles: If True then circle edges will be faceted. Defaults
                to True.
            tolerance: faceting toleranceto use when faceting cirles and
                splines. Defaults to 1e-3.
            view_plane: The plane to project. Options are 'XZ', 'XY', 'YZ',
                'YX', 'ZY', 'ZX', 'RZ' and 'XYZ'. Defaults to 'RZ'. Defaults to
                'RZ'.
        Returns:
            plotly.Figure(): figure object
        """

        fig = paramak.utils.export_wire_to_html(
            wires=self.solid.Edges(),
            filename=filename,
            view_plane=view_plane,
            facet_splines=facet_splines,
            facet_circles=facet_circles,
            tolerance=tolerance,
            title="coordinates of the " + self.__class__.__name__ +
            " reactor, viewed from the " + view_plane + " plane",
            mode="lines",
        )

        return fig

    def volume(
        self,
        split_compounds: bool = False
    ) -> List[float]:
        """Get the volumes of the Shapes in the Reactor.

        Args:
            split_compounds: If the Shape is a compound of Shapes and therefore
                contains multiple volumes. This option allows access to the separate
                volumes of each component within a Shape (True) or the volumes of
                compounds can be summed (False).

        Returns:
            The the volumes of the Shapes
        """

        all_volumes = []
        for shape in self.shapes_and_components:
            all_volumes.append(shape.volume(split_compounds=split_compounds))
        return all_volumes




import math
from typing import Optional
import numpy as np
from paramak import ExtrudeMixedShape


class InnerTfCoilsCircular(ExtrudeMixedShape):
    """A tf coil volume with cylindrical inner and outer profiles and
    constant gaps between each coil.

    Args:
        height: height of tf coils.
        inner_radius: inner radius of tf coils.
        outer_radius: outer radius of tf coils.
        number_of_coils: number of tf coils.
        gap_size: gap between adjacent tf coils.
        azimuth_start_angle: Defaults to 0.0.
        workplane: Defaults to "XY".
        rotation_axis: Defaults to "Z".
    """

    def __init__(
        self,
        height: float,
        inner_radius: float,
        outer_radius: float,
        number_of_coils: int,
        gap_size: float,
        azimuth_start_angle: Optional[float] = 0.0,
        workplane: Optional[str] = "XY",
        rotation_axis: Optional[str] = "Z",
        **kwargs
    ) -> None:

        super().__init__(
            distance=height,
            workplane=workplane,
            rotation_axis=rotation_axis,
            **kwargs
        )

        self.azimuth_start_angle = azimuth_start_angle
        self.height = height
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.number_of_coils = number_of_coils
        self.gap_size = gap_size
        self.distance = height

    @property
    def azimuth_start_angle(self):
        return self._azimuth_start_angle

    @azimuth_start_angle.setter
    def azimuth_start_angle(self, value):
        self._azimuth_start_angle = value

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def distance(self):
        return self.height

    @distance.setter
    def distance(self, value):
        self._distance = value

    @property
    def inner_radius(self):
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, inner_radius):
        self._inner_radius = inner_radius

    @property
    def outer_radius(self):
        return self._outer_radius

    @outer_radius.setter
    def outer_radius(self, outer_radius):
        self._outer_radius = outer_radius

    @property
    def number_of_coils(self):
        return self._number_of_coils

    @number_of_coils.setter
    def number_of_coils(self, number_of_coils):
        self._number_of_coils = number_of_coils

    @property
    def gap_size(self):
        return self._gap_size

    @gap_size.setter
    def gap_size(self, gap_size):
        self._gap_size = gap_size

    def find_points(self):
        """Finds the points that describe the 2D profile of the tf coil shape"""

        if self.gap_size * self.number_of_coils > 2 * math.pi * self.inner_radius:
            raise ValueError('gap_size is too large')

        theta_inner = (
            (2 * math.pi * self.inner_radius) - (self.gap_size * self.number_of_coils)
        ) / (self.inner_radius * self.number_of_coils)
        omega_inner = math.asin(self.gap_size / (2 * self.inner_radius))

        theta_outer = (
            (2 * math.pi * self.outer_radius) - (self.gap_size * self.number_of_coils)
        ) / (self.outer_radius * self.number_of_coils)
        omega_outer = math.asin(self.gap_size / (2 * self.outer_radius))

        # inner points
        point_1 = (
            (self.inner_radius * math.cos(-omega_inner)),
            (-self.inner_radius * math.sin(-omega_inner)),
        )
        point_2 = (
            (
                self.inner_radius * math.cos(theta_inner / 2) * math.cos(-omega_inner)
                + self.inner_radius * math.sin(theta_inner / 2) * math.sin(-omega_inner)
            ),
            (
                -self.inner_radius * math.cos(theta_inner / 2) * math.sin(-omega_inner)
                + self.inner_radius * math.sin(theta_inner / 2) * math.cos(-omega_inner)
            ),
        )
        point_3 = (
            (
                self.inner_radius * math.cos(theta_inner) * math.cos(-omega_inner)
                + self.inner_radius * math.sin(theta_inner) * math.sin(-omega_inner)
            ),
            (
                -self.inner_radius * math.cos(theta_inner) * math.sin(-omega_inner)
                + self.inner_radius * math.sin(theta_inner) * math.cos(-omega_inner)
            ),
        )

        # outer points
        point_4 = (
            (self.outer_radius * math.cos(-omega_outer)),
            (-self.outer_radius * math.sin(-omega_outer)),
        )
        point_5 = (
            (
                self.outer_radius * math.cos(theta_outer / 2) * math.cos(-omega_outer)
                + self.outer_radius * math.sin(theta_outer / 2) * math.sin(-omega_outer)
            ),
            (
                -self.outer_radius * math.cos(theta_outer / 2) * math.sin(-omega_outer)
                + self.outer_radius * math.sin(theta_outer / 2) * math.cos(-omega_outer)
            ),
        )
        point_6 = (
            (
                self.outer_radius * math.cos(theta_outer) * math.cos(-omega_outer)
                + self.outer_radius * math.sin(theta_outer) * math.sin(-omega_outer)
            ),
            (
                -self.outer_radius * math.cos(theta_outer) * math.sin(-omega_outer)
                + self.outer_radius * math.sin(theta_outer) * math.cos(-omega_outer)
            ),
        )

        points = [
            (point_1[0], point_1[1], "circle"),
            (point_2[0], point_2[1], "circle"),
            (point_3[0], point_3[1], "straight"),
            (point_6[0], point_6[1], "circle"),
            (point_5[0], point_5[1], "circle"),
            (point_4[0], point_4[1], "straight"),
        ]

        self.points = points

    def find_azimuth_placement_angle(self):
        """Calculates the azimuth placement angles based on the number of tf
        coils"""

        angles = list(
            np.linspace(
                0 + self.azimuth_start_angle,
                360 + self.azimuth_start_angle,
                self.number_of_coils,
                endpoint=False))

        self.azimuth_placement_angle = angles
